"""
# Normalization Layers API

This module provides classes to build normalization layers.

Functions:
    layer_normalization: Layer normalization
    batch_normalization: Batch normalization
    normalization: Normalization builder layer

Please check [Keras Normalization Layers](https://keras.io/api/layers/normalization_layers/) for additional layers.
"""

import keras


def layer_normalization(
    name: str | None = None, axis: int | tuple[int] | None = None, scale: bool = True
) -> keras.Layer:
    """Layer normalization

    If axis is None, this layer will infer based on the input tensor shape:

    * If rank is 4 (B, H, W, C), normalize over H, W
    * If rank is 3 (B, T, C), normalize over T
    * If rank is 2 (B, C), normalize over C

    Args:
        name (str | None, optional): Layer name. Defaults to None.
        axis (int | tuple[int] | None, optional): Axis. Defaults to None.
        scale (bool, optional): Scale. Defaults to True.

    Returns:
        keras.Layer: Layer
    """
    name = name + ".ln" if name else None

    if axis is None:

        def layer(x: keras.KerasTensor) -> keras.KerasTensor:
            """Functional layer normalization layer

            Args:
                x (tf.Tensor): Input tensor

            Returns:
                tf.Tensor: Output tensor
            """
            is_channels_first = keras.backend.image_data_format() == "channels_first"
            # If rank is 4 (B, H, W, C), normalize over H, W
            if x.shape.rank == 4:
                _axis = [-2, -3] if not is_channels_first else [-1, -2]
            # If rank is 3 (B, T, C), normalize over T
            elif x.shape.rank == 3:
                _axis = -2 if not is_channels_first else -1
            # If rank is 2 (B, C), normalize over C
            else:
                _axis = -1
            # END IF

            return keras.layers.LayerNormalization(axis=_axis, name=name, scale=scale)(x)

        # END DEF
    else:

        def layer(x: keras.KerasTensor) -> keras.KerasTensor:
            """Functional layer normalization layer

            Args:
                x (tf.Tensor): Input tensor

            Returns:
                tf.Tensor: Output tensor
            """
            return keras.layers.LayerNormalization(axis=axis, name=name, scale=scale)(x)

        # END DEF

    return layer


def batch_normalization(name: str | None = None, momentum=0.9, epsilon=1e-3, axis: int | None = -1) -> keras.Layer:
    """Batch normalization

    Args:
        name (str | None, optional): Layer name. Defaults to None.
        momentum (float, optional): Momentum. Defaults to 0.9.
        epsilon (float, optional): Epsilon. Defaults to 1e-3.
        axis (int | None, optional): Axis. Defaults to None.

    Returns:
        keras.Layer: Layer
    """
    name = name + "_bn" if name else None

    def layer(x: keras.KerasTensor) -> keras.KerasTensor:
        """Functional batch normalization layer

        Args:
            x (tf.Tensor): Input tensor

        Returns:
            tf.Tensor: Output tensor
        """
        return keras.layers.BatchNormalization(momentum=momentum, epsilon=epsilon, axis=axis, name=name)(x)

    # END DEF

    return layer


def normalization(norm: str, name: str, **kwargs) -> keras.Layer:
    """Creates normalization layer based on type

    Args:
        norm (str): Normalization type
        name (str): Name

    Returns:
        KerasLayer: Layer
    """
    if norm == "batch":
        return batch_normalization(name=name, **kwargs)

    if norm == "layer":
        return layer_normalization(name=name, **kwargs)

    raise ValueError(f"Normalization type '{norm}' not supported")



"""
# FLOPs Metrics API

FLOPS calculation for keras.Model or keras.Sequential.

Functions:
    get_flops: Calculate FLOPS for keras.Model or keras.Sequential


"""

import os
import keras


def get_flops(model: keras.Model, batch_size: int | None = None, fpath: os.PathLike | None = None) -> float:
    """Calculate FLOPS for keras.Model or keras.Sequential.
    Ignore operations used in only training mode such as Initialization.

    Note:
        Only tensorflow backend is supported currently.
        Does not support LSTM and GRU.

    Args:
        model (keras.Model|keras.Sequential): Model
        batch_size (int, optional): Batch size. Defaults to None.
        fpath (os.PathLike, optional): Output file path. Defaults to None.

    Returns:
        float: FLOPS

    Example:

    ```python
    model = keras.Sequential([
        keras.layers.Dense(128, activation="relu", input_shape=(784,)),
        keras.layers.Dense(64, activation="relu"),
        keras.layers.Dense(10, activation="softmax"),
    ])
    import helia_edge as helia

    model.build()
    flops = helia.metrics.get_flops(model, batch_size=1)
    print(f"FLOPS: {flops/1e6:.2f}M")
    ```
    """
    if keras.backend.backend() != "tensorflow":
        raise ValueError("Only tensorflow backend is supported.")

    import tensorflow as tf

    # pylint: disable=no-name-in-module
    from tensorflow.python.profiler.model_analyzer import profile

    # pylint: disable=no-name-in-module
    from tensorflow.python.profiler.option_builder import ProfileOptionBuilder

    input_signature = [tf.TensorSpec(shape=(batch_size,) + model.input_shape[1:])]
    forward_pass = tf.function(model.call, input_signature=input_signature)
    graph = forward_pass.get_concrete_function().graph
    options = ProfileOptionBuilder.float_operation()
    if fpath:
        options["output"] = f"file:outfile={fpath}"
    graph_info = profile(graph, options=options)
    return float(graph_info.total_float_ops)



"""
# Composer Model API

This module provides utility functions to compose a sequential set of networks/layers.

Classes:
    ComposerLayerParams: Composer layer parameters
    ComposerParams: Composer Network parameters
    ComposerModel: Helper class to generate model from parameters

Functions:
    composer_layer: Composes a sequential set of networks/layers

"""

import logging

import keras
from pydantic import BaseModel, Field

from ..layers.normalization import batch_normalization
from ..layers.convolutional import conv2d
from ..layers.squeeze_excite import se_layer
from ..layers.activations import relu6
from .utils import load_model

logger = logging.getLogger(__name__)


class ComposerLayerParams(BaseModel):
    """Composer layer parameters

    Attributes:
        name (str): Layer name
        params (dict): Layer arguments
    """

    name: str = Field(..., description="Layer name")
    params: dict = Field(default_factory=dict, description="Layer arguments")


class ComposerParams(BaseModel):
    """Composer Network parameters

    Attributes:
        layers (list[ComposerLayerParams]): Network layers
        include_top (bool): Include top
        output_activation (str | None): Output activation
        name (str): Model name
    """

    layers: list[ComposerLayerParams] = Field(default_factory=list, description="Network layers")
    include_top: bool = Field(default=True, description="Include top")
    output_activation: str | None = Field(default=None, description="Output activation")
    name: str = Field(default="Composer", description="Model name")


def composer_layer(
    x: keras.KerasTensor,
    params: ComposerParams,
    num_classes: int | None = None,
) -> keras.KerasTensor:
    """Composes a sequential set of networks/layers.
    Useful for adding custom layers to a pre-trained model (e.g. foundation).

    Args:
        x (keras.KerasTensor): Model input
        params (ComposerParams): Model parameters
        num_classes (int | None): Number of classes

    Returns:
        keras.KerasTensor: Model output
    """
    y = x
    for layer in params.layers:
        match layer.name:
            case "conv2d":
                y = conv2d(y, **layer.params)
            case "dense":
                y = keras.layers.Dense(**layer.params)(y)
            case "relu6":
                y = relu6(y)
            case "batch_norm":
                y = batch_normalization(y)
            case "se_block":
                y = se_layer(y, **layer.params)
            case "load_model":
                prev_model = load_model(layer.params["model_file"])
                trainable = layer.params.get("trainable", True)
                if not trainable:
                    logger.info(f"Freezing model {prev_model.name}")
                prev_model.trainable = trainable
                y = prev_model(y, training=trainable)
            case _:
                raise ValueError(f"Unknown layer {layer.name}")
        # END MATCH
    # END FOR

    if params.include_top:
        if num_classes is not None:
            y = keras.layers.Dense(num_classes)(y)
        if params.output_activation:
            y = keras.layers.Activation(params.output_activation)(y)

    return y


class ComposerModel:
    """Helper class to generate model from parameters"""

    @staticmethod
    def layer_from_params(inputs: keras.Input, params: ComposerParams | dict, num_classes: int | None = None):
        """Create layer from parameters"""
        if isinstance(params, dict):
            params = ComposerParams(**params)
        return composer_layer(x=inputs, params=params, num_classes=num_classes)

    @staticmethod
    def model_from_params(inputs: keras.Input, params: ComposerParams | dict, num_classes: int | None = None):
        """Create model from parameters"""
        outputs = ComposerModel.layer_from_params(inputs=inputs, params=params, num_classes=num_classes)
        return keras.Model(inputs=inputs, outputs=outputs)



"""
# Layer Normalization Layer API

This module provides classes to build layer normalization layers.

Classes:
    LayerNormalization1D: Layer normalization for 1D data
    LayerNormalization2D: Layer normalization for 2D data

"""

import keras
from .base_augmentation import BaseAugmentation1D, BaseAugmentation2D
from ...utils import helia_export


@helia_export(path="helia_edge.layers.preprocessing.LayerNormalization1D")
class LayerNormalization1D(BaseAugmentation1D):
    def __init__(
        self,
        epsilon: float = 1e-6,
        name=None,
        **kwargs,
    ):
        """Apply Layer Normalization to the input.

        Args:
            epsilon (float): Small value to avoid division by zero.
        """
        super().__init__(name=name, **kwargs)
        self.epsilon = epsilon

    def augment_samples(self, inputs) -> keras.KerasTensor:
        """Augment a batch of samples during training."""

        samples = inputs[self.SAMPLES]

        mean = keras.ops.mean(samples, axis=self.data_axis, keepdims=True)
        variance = keras.ops.var(samples, axis=self.data_axis, keepdims=True)
        outputs = (samples - mean) / keras.ops.sqrt(variance + self.epsilon)

        return outputs  # (batch, duration, channels)


@helia_export(path="helia_edge.layers.preprocessing.LayerNormalization2D")
class LayerNormalization2D(BaseAugmentation2D):
    def __init__(
        self,
        epsilon: float = 1e-6,
        name=None,
        **kwargs,
    ):
        """Apply Layer Normalization to the input.

        Args:
            epsilon (float): Small value to avoid division by zero.
        """
        super().__init__(name=name, **kwargs)
        self.epsilon = epsilon

    def augment_samples(self, inputs) -> keras.KerasTensor:
        """Augment a batch of samples during training."""

        samples = inputs[self.SAMPLES]
        axis = (self.height_axis, self.width_axis)
        mean = keras.ops.mean(samples, axis=axis, keepdims=True)
        variance = keras.ops.var(samples, axis=axis, keepdims=True)
        outputs = (samples - mean) / keras.ops.sqrt(variance + self.epsilon)

        return outputs  # (batch, duration, channels)



"""
# Masked Autoencoder Trainer API

This module contains the implementation of a masked autoencoder trainer that
can be used to train a model using the masked autoencoder approach.

Classes:
    MaskedAutoencoder: A trainer for masked autoencoder

"""

from typing import Callable

import keras
import tensorflow as tf

from ..utils import helia_export


@helia_export(path="helia_edge.trainers.MaskedAutoencoder")
class MaskedAutoencoder(keras.Model):
    def __init__(
        self,
        patch_layer: Callable[
            [keras.KerasTensor],
            tuple[
                keras.KerasTensor,
                keras.KerasTensor,
                keras.KerasTensor,
                keras.KerasTensor,
                keras.KerasTensor,
            ],
        ],
        patch_encoder: Callable[[keras.KerasTensor], keras.KerasTensor],
        encoder: keras.Model,
        decoder: keras.Model,
        **kwargs,
    ):
        """Masked Autoencoder model for self-supervised learning.

        Args:
            patch_layer (Callable[[keras.KerasTensor], tuple[keras.KerasTensor, keras.KerasTensor, keras.KerasTensor, keras.KerasTensor, keras.KerasTensor]]): The patch layer which will extract patches from the input.
            patch_encoder (Callable[[keras.KerasTensor], keras.KerasTensor]): The patch encoder which will encode the patches.
            encoder (keras.Model): The encoder model.
            decoder (keras.Model): The decoder model.
        """
        super().__init__(**kwargs)
        self.patch_layer = patch_layer
        self.patch_encoder = patch_encoder
        self.encoder = encoder
        self.decoder = decoder

    def calculate_loss(self, x: keras.KerasTensor, test: bool = False):
        """Calculate the loss for the Masked Autoencoder model.

        Args:
            x (keras.KerasTensor): The input tensor.
            test (bool, optional): Whether the model is testing. Defaults to False.
        """
        # Patch the input.
        patches = self.patch_layer(x)

        # Encode the patches.
        (
            unmasked_embeddings,
            masked_embeddings,
            unmasked_positions,
            mask_indices,
            unmask_indices,
        ) = self.patch_encoder(patches)

        # Pass the unmasked patches to the encoder.
        encoder_outputs = self.encoder(unmasked_embeddings)

        # Create the decoder inputs.
        encoder_outputs = encoder_outputs + unmasked_positions
        decoder_inputs = keras.ops.concatenate([encoder_outputs, masked_embeddings], axis=1)

        # Decode the inputs.
        decoder_outputs = self.decoder(decoder_inputs)
        decoder_patches = self.patch_layer(decoder_outputs)

        loss_patch = tf.gather(patches, mask_indices, axis=1, batch_dims=1)
        loss_output = tf.gather(decoder_patches, mask_indices, axis=1, batch_dims=1)

        # Compute the total loss.
        total_loss = self.compute_loss(y=loss_patch, y_pred=loss_output)

        return total_loss, loss_patch, loss_output

    def _tensorflow_train_step(self, x):
        with tf.GradientTape() as tape:
            total_loss, loss_patch, loss_output = self.calculate_loss(x)

        # Apply gradients.
        train_vars = [
            self.patch_layer.trainable_variables,
            self.patch_encoder.trainable_variables,
            self.encoder.trainable_variables,
            self.decoder.trainable_variables,
        ]
        grads = tape.gradient(total_loss, train_vars)
        tv_list = []
        for grad, var in zip(grads, train_vars):
            for g, v in zip(grad, var):
                tv_list.append((g, v))
        self.optimizer.apply_gradients(tv_list)

        # Report progress.
        results = {}
        for metric in self.metrics:
            metric.update_state(loss_patch, loss_output)
            results[metric.name] = metric.result()
        return results

    def _tensorflow_test_step(self, x):
        total_loss, loss_patch, loss_output = self.calculate_loss(x, test=True)

        # Update the trackers.
        results = {}
        for metric in self.metrics:
            metric.update_state(loss_patch, loss_output)
            results[metric.name] = metric.result()
        return results

    def train_step(self, *args, **kwargs):
        if keras.backend.backend() == "tensorflow":
            return self._tensorflow_train_step(*args, **kwargs)
        elif keras.backend.backend() == "jax":
            raise NotImplementedError("JAX backend is not supported.")
        elif keras.backend.backend() == "torch":
            raise NotImplementedError("PyTorch backend is not supported.")

    def test_step(self, *args, **kwargs):
        if keras.backend.backend() == "tensorflow":
            return self._tensorflow_test_step(*args, **kwargs)
        elif keras.backend.backend() == "jax":
            raise NotImplementedError("JAX backend is not supported.")
        elif keras.backend.backend() == "torch":
            raise NotImplementedError("PyTorch backend is not supported.")



from typing import TypeAlias, Mapping, List
import keras

NestedTensorType: TypeAlias = List["NestedTensorValue"] | Mapping[str, "NestedTensorValue"]
NestedTensorValue: TypeAlias = keras.KerasTensor | NestedTensorType



"""
# Environment Utility API

This module provides utility functions to interact with the environment.

Functions:
    setup_logger: Setup logger with Rich
    env_flag: Return the specified environment variable coerced to a bool
    silence_tensorflow: Silence every unnecessary warning from tensorflow
    disable_tensorflow_gpu: Disable TensorFlow GPU
"""

import os
import sys
from contextlib import contextmanager
import logging

from rich.logging import RichHandler


def setup_logger(
    log_name: str,
    level: int | None = None,
    recursive: bool = True,
    file_path: os.PathLike | None = None,
) -> logging.Logger:
    """Setup logger with Rich

    Args:
        log_name (str): Logger name

    Returns:
        logging.Logger: Logger
    """
    root_name = log_name.split(".")[0] if recursive else log_name
    new_logger = logging.getLogger(root_name)
    needs_init = not new_logger.handlers

    match level:
        case 0:
            log_level = logging.ERROR
        case 1:
            log_level = logging.INFO
        case 2 | 3 | 4:
            log_level = logging.DEBUG
        case None:
            log_level = None
        case _:
            log_level = logging.INFO
    # END MATCH

    if needs_init:
        logging.basicConfig(force=True, handlers=[RichHandler(rich_tracebacks=True)])
        new_logger.propagate = False
        new_logger.handlers = [RichHandler(show_time=False)]

    if log_level is not None:
        new_logger.setLevel(log_level)

    if file_path and not any(isinstance(handler, logging.FileHandler) for handler in new_logger.handlers):
        file_handler = logging.FileHandler(file_path, mode="w")
        file_handler.setLevel(log_level)
        new_logger.addHandler(file_handler)

    return new_logger


def env_flag(env_var: str, default: bool = False) -> bool:
    """Return the specified environment variable coerced to a bool, as follows:
    - When the variable is unset, or set to the empty string, return `default`.
    - When the variable is set to a truthy value, returns `True`.
      These are the truthy values:
          - 1
          - true, yes, on
    - When the variable is set to the anything else, returns False.
       Example falsy values:
          - 0
          - no
    - Ignore case and leading/trailing whitespace.

    Args:
        env_var (str): Environment variable name
        default (bool, optional): Default value. Defaults to False.

    Returns:
        bool: Value of environment variable
    """
    environ_string = os.environ.get(env_var, "").strip().lower()
    if not environ_string:
        return default
    return environ_string in ["1", "true", "yes", "on"]


def silence_tensorflow():
    """Silence every unnecessary warning from tensorflow."""
    logging.getLogger("tensorflow").setLevel(logging.ERROR)
    os.environ["KMP_AFFINITY"] = "noverbose"
    os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
    os.environ["AUTOGRAPH_VERBOSITY"] = "5"
    # We wrap this inside a try-except block
    # because we do not want to be the one package
    # that crashes when TensorFlow is not installed
    # when we are the only package that requires it
    # in a given Jupyter Notebook, such as when the
    # package import is simply copy-pasted.
    try:
        import tensorflow as tf

        tf.get_logger().setLevel("ERROR")
        tf.autograph.set_verbosity(3)
    except ModuleNotFoundError:
        pass


def disable_tensorflow_gpu():
    """Disable TensorFlow GPU"""
    import tensorflow as tf

    os.environ["CUDA_VISIBLE_DEVICES"] = "-1"
    tf.config.set_visible_devices([], "GPU")


@contextmanager
def suppress_os_stdio():
    """
    Redirects the OS stdout (fd=1) and stderr (fd=2) to /dev/null, then restores them on exit.
    Suppresses all output **unless** an exception is raised, in which case it will be logged.

    Example:
        with suppress_os_stdio():
            noisy_function()
    """
    # 1) Flush Python buffers
    sys.stdout.flush()
    sys.stderr.flush()

    # 2) Duplicate original FDs so we can restore them later
    orig_stdout_fd = os.dup(sys.stdout.fileno())
    orig_stderr_fd = os.dup(sys.stderr.fileno())

    try:
        # 3) Open /dev/null and redirect
        devnull_fd = os.open(os.devnull, os.O_RDWR)
        os.dup2(devnull_fd, sys.stdout.fileno())
        os.dup2(devnull_fd, sys.stderr.fileno())
        os.close(devnull_fd)

        # 4) Enter user’s block
        yield

    except Exception:
        # 5) An error occurred! Log it (the FDs are still /dev/null at this moment).
        logging.exception("Error inside suppressed stdout/stderr block")
        raise

    finally:
        # 6) Restore original FDs no matter what
        os.dup2(orig_stdout_fd, sys.stdout.fileno())
        os.dup2(orig_stderr_fd, sys.stderr.fileno())
        os.close(orig_stdout_fd)
        os.close(orig_stderr_fd)



"""
# Tensor Utility API

This module provides utility functions for working with tensors.

Functions:
    matches_spec: Test whether data object matches the desired spec

"""

import numpy.typing as npt

SpecType = list[npt.ArrayLike] | tuple[npt.ArrayLike] | dict[str, npt.ArrayLike] | npt.ArrayLike


def matches_spec(o: SpecType, spec: SpecType, ignore_batch_dim: bool = False) -> bool:
    """Test whether data object matches the desired spec.

    Args:
        o (SpecType): Data object.
        spec (SpecType): Metadata for describing the the data object.
        ignore_batch_dim: Ignore first dimension when checking shapes.

    Returns:
        bool: If matches
    """
    if isinstance(spec, (list, tuple)):
        if not isinstance(o, (list, tuple)):
            raise ValueError(f"data object is not a list or tuple which is required by the spec: {spec}")
        if len(spec) != len(o):
            raise ValueError(f"data object has a different number of elements than the spec: {spec}")
        for i, ispec in enumerate(spec):
            if not matches_spec(o[i], ispec, ignore_batch_dim=ignore_batch_dim):
                return False
        return True

    if isinstance(spec, dict):
        if not isinstance(o, dict):
            raise ValueError(f"data object is not a dict which is required by the spec: {spec}")
        if spec.keys() != o.keys():
            raise ValueError(f"data object has different keys than those specified in the spec: {spec}")
        for k in spec:
            if not matches_spec(o[k], spec[k], ignore_batch_dim=ignore_batch_dim):
                return False
            return True

    spec_shape = spec.shape[1:] if ignore_batch_dim else spec.shape
    o_shape = o.shape[1:] if ignore_batch_dim else o.shape
    return spec_shape == o_shape and spec.dtype == o.dtype



"""
# Augmentation Pipeline API

Classes:
    AugmentationPipeline: Pipeline of augmentation layers
"""

import keras

from ...utils import helia_export


@helia_export(path="helia_edge.layers.preprocessing.AugmentationPipeline")
class AugmentationPipeline(keras.Layer):
    def __init__(
        self,
        layers: list[keras.Layer],
        name: str | None = None,
        force_training: bool = False,
    ):
        """Pipeline of augmentation layers.

        Args:
            layers (list[keras.Layer]): List of augmentation layers.
            force_training (bool, optional): Force training mode. Defaults to False.

        Example:

        ```python
        import helia_edge as helia

        layers = [
            helia.layers.preprocessing.RandomNoiseDistortion1D(sample_rate=100, frequency=(1, 2), amplitude=(0.5, 2)),
            helia.layers.preprocessing.AmplitudeWarp(sample_rate=100, frequency=(1, 2), amplitude=(0.5, 2)),
        ]
        pipeline = helia.layers.preprocessing.AugmentationPipeline(layers)
        x = keras.random.normal((10, 100, 1), dtype="float32")
        x_aug = pipeline(x, training=True)
        plt.plot(x[0].numpy())
        plt.plot(x_aug[0].numpy())
        plt.show()
        ```
        """
        super().__init__(name=name)
        self.layers = layers
        self.force_training = force_training

    def call(self, inputs, training: bool = True, **kwargs):
        for layer in self.layers:
            inputs = layer(inputs, training=training or self.force_training, **kwargs)
        return inputs



"""
# ConvMixer

For more info, refer to the original paper [ConvMixer: Revisiting Convolution in Vision](https://arxiv.org/abs/2201.09792).

Classes:
    ConvMixerParams: ConvMixer parameters
    ConvMixerModel: Helper class to generate model from parameters

Functions:
    conv_mixer_block: ConvMixer block
    conv_mixer_layer: ConvMixer layer

"""

from typing import Callable

import keras
from pydantic import BaseModel, Field

from ..layers.normalization import batch_normalization


class ConvMixerParams(BaseModel):
    """ConvMixer parameters

    Attributes:
        filters (int): Number of filters per layer
        depth (int): Network depth
        kernel_size (int): Filter size
        patch_size (int): Patch size
        include_top (bool): Include top
        output_activation (str | None): Output activation
        name (str): Model name

    """

    filters: int = Field(default=256, description="# filters per layer")
    depth: int = Field(default=8, description="Network depth")
    kernel_size: int = Field(default=5, description="Filter size")
    patch_size: int = Field(default=2, description="Patch size")
    include_top: bool = Field(default=True, description="Include top")
    output_activation: str | None = Field(default=None, description="Output activation")
    name: str = Field(default="RegNet", description="Model name")


def conv_mixer_block(filters: int, kernel_size: int) -> Callable[[keras.KerasTensor], keras.KerasTensor]:
    """ConvMixer block

    Args:
        filters (int): Number of filters
        kernel_size (int): Kernel size
    """

    def layer(x: keras.KerasTensor) -> keras.KerasTensor:
        # Depthwise convolution.
        x0 = x
        x = keras.layers.DepthwiseConv2D(kernel_size=kernel_size, padding="same")(x)
        x = keras.layers.Activation("gelu")(x)
        x = batch_normalization()(x)
        # Residual
        x = keras.layers.Add()([x, x0])

        # Pointwise convolution.
        x = keras.layers.Conv2D(filters, kernel_size=1)(x)
        x = keras.layers.Activation("gelu")(x)
        x = batch_normalization()(x)
        return x

    return layer


def conv_mixer_layer(
    x: keras.KerasTensor,
    params: ConvMixerParams,
    num_classes: int | None = None,
) -> keras.KerasTensor:
    """ConvMixer: https://openreview.net/pdf?id=TVHS5Y4dNvM.

    Args:
        x (keras.KerasTensor): Input tensor
        params (ConvMixerParams): Model parameters.
        num_classes (int, optional): Number of classes.

    Returns:
        keras.KerasTensor: Model output
    """
    # Extract patch embeddings
    y = keras.layers.Conv2D(
        filters=params.filters,
        kernel_size=params.patch_size,
        strides=params.patch_size,
        padding="same",
        use_bias=True,
    )(x)
    y = keras.layers.Activation("gelu")(y)
    y = batch_normalization()(y)

    # ConvMixer blocks
    for _ in range(params.depth):
        y = conv_mixer_block(filters=params.filters, kernel_size=params.kernel_size)(y)

    # Classification block
    if params.include_top:
        y = keras.layers.GlobalAvgPool2D(keepdims=False)(y)
        if num_classes is not None:
            y = keras.layers.Dense(num_classes)(y)
        if params.output_activation:
            y = keras.layers.Activation(params.output_activation)(y)

    return y


class ConvMixerModel:
    """Helper class to generate model from parameters"""

    @staticmethod
    def layer_from_params(inputs: keras.Input, params: ConvMixerParams | dict, num_classes: int | None = None):
        """Create layer from parameters"""
        if isinstance(params, dict):
            params = ConvMixerParams(**params)
        return conv_mixer_layer(x=inputs, params=params, num_classes=num_classes)

    @staticmethod
    def model_from_params(inputs: keras.Input, params: ConvMixerParams | dict, num_classes: int | None = None):
        """Create model from parameters"""
        outputs = ConvMixerModel.layer_from_params(inputs=inputs, params=params, num_classes=num_classes)
        return keras.Model(inputs=inputs, outputs=outputs)



"""
# TFLite Interpreter API

This module handles interpreting TensorFlow Lite models.

Classes:
    TfLiteKerasInterpreter: Interprets TensorFlow Lite models converted from Keras.

"""

from .interpreter import TfLiteKerasInterpreter



"""
# Resizing layers API

This module provides classes to build resizing layers.

Classes:
    Resizing1D: Resize 1D samples
    Resizing2D: Resize 2D samples
"""

import keras
from .base_augmentation import BaseAugmentation1D, BaseAugmentation2D
from ...utils import helia_export


@helia_export(path="helia_edge.layers.preprocessing.Resizing1D")
class Resizing1D(BaseAugmentation1D):
    duration: int
    data_format: str

    def __init__(self, duration: int, **kwargs):
        """1D resizing layer

        Args:
            duration (int): The new duration of the samples

        """
        super().__init__(**kwargs)
        self.duration = duration

    def augment_samples(self, inputs) -> keras.KerasTensor:
        """Resize a batch of samples during training."""
        samples = inputs[self.SAMPLES]

        # Add height dimension
        samples = keras.ops.expand_dims(samples, axis=1)
        samples = keras.ops.image.resize(
            samples,
            size=(1, self.duration),
            interpolation="bicubic",
            crop_to_aspect_ratio=False,
            data_format=self.data_format,
        )
        # Remove height dimension
        samples = keras.ops.squeeze(samples, axis=1)
        return samples

    def compute_output_shape(self, input_shape, *args, **kwargs):
        """Compute output shape."""
        output_shape = list(input_shape)
        output_shape[self.data_axis] = self.duration
        return tuple(output_shape)

    def get_config(self):
        """Serialize the configuration."""
        config = super().get_config()
        config.update(
            duration=self.duration,
            data_format=self.data_format,
        )
        return config


@helia_export(path="helia_edge.layers.preprocessing.Resizing2D")
class Resizing2D(BaseAugmentation2D):
    height: int
    width: int
    interpolation: str

    def __init__(self, height: int, width: int, interpolation: str = "bicubic", **kwargs):
        """"""
        super().__init__(**kwargs)
        self.height = height
        self.width = width

    def augment_samples(self, inputs) -> keras.KerasTensor:
        """Resize a batch of samples during training."""
        samples = inputs[self.SAMPLES]
        samples = keras.ops.image.resize(
            samples,
            size=(self.height, self.width),
            interpolation="bicubic",
            crop_to_aspect_ratio=False,
            data_format=self.data_format,
        )
        return samples

    def compute_output_shape(self, input_shape, *args, **kwargs):
        """Compute output shape."""
        output_shape = list(input_shape)
        output_shape[self.height_axis] = self.height
        output_shape[self.width_axis] = self.width
        return tuple(output_shape)

    def get_config(self):
        """Serialize the configuration."""
        config = super().get_config()
        config.update(
            height=self.height,
            width=self.width,
        )
        return config



"""
# Random Cutout Layer API

This module provides classes to build random cutout layers.

Classes:
    RandomCutout1D: Random cutout 1D
    RandomCutout2D: Random cutout 2D

"""

import keras

from ...utils import parse_factor
from .base_augmentation import BaseAugmentation1D, BaseAugmentation2D
from ...utils import helia_export


@helia_export(path="helia_edge.layers.preprocessing.RandomCutout1D")
class RandomCutout1D(BaseAugmentation1D):
    cutouts: int
    factor: tuple[float, float]
    fill_mode: str
    fill_value: float

    def __init__(
        self,
        factor: float | tuple[float, float] = 0.1,
        cutouts: int = 1,
        fill_mode="constant",
        fill_value: float = 0.0,
        **kwargs,
    ):
        """Apply random cutout to the input. This is similar to its 2D counterpart where a random portion of the input is cutout.
        We allow providing a range for the factor and cutouts to randomly pick the values.

        Args:
            factor (float|tuple[float,float]): Factor of the duration to cutout. If tuple, factor is randomly picked between the values.
            cutouts (int): Number of cutouts to apply.
            fill_mode (str): Fill mode. "constant" or "normal".
            fill_value (float): Fill value for the cutout.

        """
        super().__init__(**kwargs)

        self.factor = parse_factor(factor, min_value=0, max_value=1, param_name="factor")
        self.cutouts = cutouts
        self.fill_mode = fill_mode
        self.fill_value = fill_value
        if fill_mode not in ["normal", "constant"]:
            raise ValueError(f'`fill_mode` should be "normal" or "constant". Got `fill_mode`={fill_mode}')
        # END IF

    def call(self, inputs, training: bool = True):
        """Override the call method to apply multiple cutouts."""
        self.training = training
        if not self.training:
            return inputs
        inputs, metadata = self._format_inputs(inputs)
        outputs = keras.ops.fori_loop(
            lower=0,
            upper=self.cutouts,
            body_fun=lambda _, x: self.batch_augment(x),
            init_val=self.batch_augment(inputs),
        )
        return self._format_outputs(outputs, metadata)

    def get_random_transformations(self, input_shape):
        """Generate random cutout locations, sizes, and fill values."""
        batch_size = input_shape[0]
        duration_size = input_shape[self.data_axis]

        cut_size = keras.random.randint(
            shape=(batch_size,),
            minval=int(duration_size * self.factor[0]),
            maxval=int(duration_size * self.factor[1]) + 1,
            dtype="int32",
            seed=self.random_generator,
        )
        cut_start = keras.random.randint(
            shape=(batch_size,),
            minval=0,
            maxval=int(duration_size * (1 - self.factor[1]) + 1),
            dtype="int32",
            seed=self.random_generator,
        )
        if self.fill_mode == "constant":
            fill = keras.ops.ones(input_shape) * self.fill_value
        else:
            fill = keras.random.normal(input_shape, mean=0, stddev=self.fill_value, seed=self.random_generator)

        return {
            "cut_start": cut_start,
            "cut_size": cut_size,
            "fill": fill,
        }

    def augment_sample(self, inputs) -> keras.KerasTensor:
        """Apply cutout to the input."""
        sample = inputs[self.SAMPLES]
        transforms = inputs[self.TRANSFORMS]
        cut_start = transforms["cut_start"]
        cut_size = transforms["cut_size"]
        fill = transforms["fill"]

        duration_size = sample.shape[self.data_axis]
        ch_size = sample.shape[self.ch_axis]

        if self.data_format == "channels_first":
            reshape_size = (1, duration_size)
            tile_size = (ch_size, 1)
        else:
            reshape_size = (duration_size, 1)
            tile_size = (1, ch_size)

        mask = keras.ops.tile(
            keras.ops.reshape(
                keras.ops.logical_and(
                    keras.ops.arange(duration_size) >= cut_start, keras.ops.arange(duration_size) < cut_start + cut_size
                ),
                reshape_size,
            ),
            tile_size,
        )
        result = keras.ops.where(mask, fill, sample)
        return result

    def get_config(self):
        """Serialize the configuration of the layer."""
        config = super().get_config()
        config.update(
            {
                "factor": self.factor,
                "cutouts": self.cutouts,
                "fill_mode": self.fill_mode,
                "fill_value": self.fill_value,
            }
        )
        return config


@helia_export(path="helia_edge.layers.preprocessing.RandomCutout2D")
class RandomCutout2D(BaseAugmentation2D):
    cutouts: int
    factor: tuple[float, float]
    fill_mode: str
    fill_value: float

    def __init__(
        self,
        factor: float | tuple[float, float] = 0.1,
        cutouts: int = 1,
        fill_mode="constant",
        fill_value: float = 0.0,
        **kwargs,
    ):
        """Apply random cutout to the input. This is similar to its 1D counterpart where a random portion of the input is cutout.
        We allow providing a range for the factor and cutouts to randomly pick the values.

        Args:
            factor (float|tuple[float,float]): Factor of the dimensions to cutout. If tuple, factor is randomly picked between the values.
            cutouts (int): Number of cutouts to apply.
            fill_mode (str): Fill mode. "constant" or "normal".
            fill_value (float): Fill value for the cutout.

        """
        super().__init__(**kwargs)

        self.factor = parse_factor(factor, min_value=0, max_value=1, param_name="factor")
        self.cutouts = cutouts
        self.fill_mode = fill_mode
        self.fill_value = fill_value
        if fill_mode not in ["normal", "constant"]:
            raise ValueError(f'`fill_mode` should be "normal" or "constant". Got `fill_mode`={fill_mode}')
        # END IF

    def call(self, inputs, training: bool = True):
        """Override the call method to apply multiple cutouts."""
        self.training = training
        if not self.training:
            return inputs
        inputs, metadata = self._format_inputs(inputs)
        outputs = keras.ops.fori_loop(
            lower=0,
            upper=self.cutouts,
            body_fun=lambda _, x: self.batch_augment(x),
            init_val=self.batch_augment(inputs),
        )
        return self._format_outputs(outputs, metadata)

    def get_random_transformations(self, input_shape):
        """Generate random cutout locations, sizes, and fill values."""
        batch_size = input_shape[0]
        height_size = input_shape[self.height_axis]
        width_size = input_shape[self.width_axis]

        cut_height = keras.random.randint(
            shape=(batch_size,),
            minval=int(height_size * self.factor[0]),
            maxval=int(height_size * self.factor[1]) + 1,
            dtype="int32",
            seed=self.random_generator,
        )
        cut_width = keras.random.randint(
            shape=(batch_size,),
            minval=int(width_size * self.factor[0]),
            maxval=int(width_size * self.factor[1]) + 1,
            dtype="int32",
            seed=self.random_generator,
        )
        cut_start_height = keras.random.randint(
            shape=(batch_size,),
            minval=0,
            maxval=int(height_size * (1 - self.factor[1]) + 1),
            dtype="int32",
            seed=self.random_generator,
        )
        cut_start_width = keras.random.randint(
            shape=(batch_size,),
            minval=0,
            maxval=int(width_size * (1 - self.factor[1]) + 1),
            dtype="int32",
            seed=self.random_generator,
        )
        if self.fill_mode == "constant":
            fill = keras.ops.ones(input_shape) * self.fill_value
        else:
            fill = keras.random.normal(input_shape, mean=0, stddev=self.fill_value, seed=self.random_generator)

        return {
            "cut_start_height": cut_start_height,
            "cut_start_width": cut_start_width,
            "cut_height": cut_height,
            "cut_width": cut_width,
            "fill": fill,
        }

    def augment_sample(self, inputs) -> keras.KerasTensor:
        """Apply cutout to the input."""
        sample = inputs[self.SAMPLES]
        transforms = inputs[self.TRANSFORMS]
        cut_start_height = transforms["cut_start_height"]
        cut_start_width = transforms["cut_start_width"]
        cut_height = transforms["cut_height"]
        cut_width = transforms["cut_width"]
        fill = transforms["fill"]

        height_size = sample.shape[self.height_axis]
        width_size = sample.shape[self.width_axis]
        ch_size = sample.shape[self.ch_axis]

        if self.data_format == "channels_first":
            reshape_size = (1, height_size, width_size)
            tile_size = (ch_size, 1, 1)
        else:
            reshape_size = (height_size, width_size, 1)
            tile_size = (1, 1, ch_size)

        mask = keras.ops.tile(
            keras.ops.reshape(
                keras.ops.logical_and(
                    keras.ops.logical_and(
                        keras.ops.arange(height_size) >= cut_start_height,
                        keras.ops.arange(height_size) < cut_start_height + cut_height,
                    ),
                    keras.ops.logical_and(
                        keras.ops.arange(width_size) >= cut_start_width,
                        keras.ops.arange(width_size) < cut_start_width + cut_width,
                    ),
                ),
                reshape_size,
            ),
            tile_size,
        )
        result = keras.ops.where(mask, fill, sample)
        return result

    def get_config(self):
        """Serialize the configuration of the layer."""
        config = super().get_config()
        config.update(
            {
                "factor": self.factor,
                "cutouts": self.cutouts,
                "fill_mode": self.fill_mode,
                "fill_value": self.fill_value,
            }
        )
        return config



"""
# C++ Converter API

This module handles converting models to C++ format.

Functions:
    array_dump: Generate C array of values from flattened numpy array.
    xxd_c_dump: Generate C array of values from binary file.

"""

from .utils import array_dump, xxd_c_dump



"""
# FIR Filter Layer API

This module provides classes to build FIR filter layers.

Classes:
    FirFilter: FIR filter layer

"""

import numpy as np
import numpy.typing as npt
import keras

from .base_augmentation import BaseAugmentation1D
from ...utils import helia_export


@helia_export(path="helia_edge.layers.preprocessing.FirFilter")
class FirFilter(BaseAugmentation1D):
    def __init__(
        self,
        b: npt.NDArray[np.float32],
        a: npt.NDArray[np.float32] | None = None,
        forward_backward: bool = False,
        **kwargs,
    ):
        """Apply FIR filter to the input.

        Args:
            b (np.ndarray): Numerator coefficients.
            a (np.ndarray|None): Denominator coefficients. Defaults to None.
            forward_backward (bool): Apply filter forward and backward. Defaults to False.

        Example:

        ```python
        # Create sine wave at 10 Hz with 1000 Hz sampling rate
        t = np.linspace(0, 1, 1000, endpoint=False)
        x = np.sin(2 * np.pi * 10 * t)
        # Add noise at 100 Hz and 2 Hz
        x_noise = x + 0.5 * np.sin(2 * np.pi * 100 * t) + 0.5 * np.sin(2 * np.pi * 2 * t)
        x_noise = x_noise.reshape(-1, 1).astype(np.float32)
        x_noise = keras.ops.convert_to_tensor(x_noise)
        import helia_edge as helia

        # Create bandpass FIR taps
        b = scipy.signal.firwin(101, [5, 15], fs=1000, pass_zero=False)
        lyr = helia.layers.preprocessing.FirFilter(b, forward_backward=True)
        y = lyr(x_noise).numpy().squeeze()
        x_noise = x_noise.numpy().squeeze()
        plt.plot(x, label="Original")
        plt.plot(x_noise, label="Noisy")
        plt.plot(y, label="Filtered")
        plt.legend()
        plt.show()
        ```
        """
        super().__init__(**kwargs)
        b = b.reshape(-1, 1, 1)
        self.b = self.add_weight(
            name="b",
            shape=b.shape,
            trainable=False,
        )
        self.b.assign(b)
        if a is not None:
            a = a.reshape(-1, 1, 1)
            self.a = self.add_weight(
                name="a",
                shape=a.shape,
                trainable=False,
            )
            self.a.assign(a)
        else:
            self.a = None

        self.forward_backward = forward_backward

    def augment_samples(self, inputs) -> keras.KerasTensor:
        """Augment a batch of samples during training."""

        samples = inputs[self.SAMPLES]

        # TODO: Handle 'a' coefficients
        if self.a is not None:
            raise NotImplementedError("Denominator coefficients 'a' are not supported yet.")

        outputs = keras.ops.depthwise_conv(samples, self.b, padding="same")

        if self.forward_backward:
            outputs = keras.ops.flip(outputs, axis=self.data_axis)
            outputs = keras.ops.depthwise_conv(outputs, self.b, padding="same")
            outputs = keras.ops.flip(outputs, axis=self.data_axis)
        # END IF
        return outputs

    def get_config(self):
        """Serialize the layer configuration."""
        config = super().get_config()
        config.update(
            {
                "forward_backward": self.forward_backward,
            }
        )
        return config



"""
# :material-chart-bar: Plotting Module API

This module provides utility functions to plot metrics and confusion matrices.

Functions:
    confusion_matrix_plot: Plot confusion matrix
    px_plot_confusion_matrix: Plot confusion matrix using Plotly
    multilabel_confusion_matrix: Compute multilabel confusion matrix
    multilabel_confusion_matrix_plot: Plot multilabel confusion matrix
    roc_auc_plot: Plot ROC-AUC curve
    plot_history_metrics: Plot training history metrics

"""

from .cm import (
    confusion_matrix_plot,
    px_plot_confusion_matrix,
    multilabel_confusion_matrix,
    multilabel_confusion_matrix_plot,
)
from .roc import roc_auc_plot
from .history import plot_history_metrics



"""
# C++ Converter Utilities

This module provides utility functions to convert numpy arrays and binary files to C++ format.

Functions:
    array_dump: Generate C array of values from flattened numpy array.
    xxd_c_dump: Generate C array of values from binary file.

"""

import os

import numpy as np
import numpy.typing as npt


def array_dump(
    data: npt.NDArray,
    dst_path: os.PathLike,
    var_name: str = "test_stimulus",
    var_dtype: str | None = None,
    row_len: int = 12,
    is_header: bool = False,
):
    """Generate C array of values from flattened numpy array.

    Args:
        data (npt.NDArray): Data array
        dst_path (PathLike): C file destination path
        var_name (str, optional): C variable name. Defaults to "test_stimulus".
        var_dtype (str | None, optional): C variable type. Defaults to None.
        row_len (int, optional): Elements to write per row. Defaults to 12.
        is_header (bool): Write as header or source C file. Defaults to source.

    Example:

    ```python
    import helia_edge as helia

    data = np.random.rand(10, 10).astype(np.float32)
    dst_path = "test_stimulus.h"
    helia.converters.cpp.array_dump(
        data,
        dst_path,
        var_name="test_stimulus",
        var_dtype="float32_t",
        row_len=12,
        is_header=True
    )
    ```
    """
    data = data.flatten()

    if var_dtype is not None:
        pass
    elif isinstance(data[0], np.floating):
        var_dtype = "float"
    elif isinstance(data[0], np.int8):
        var_dtype = "int8_t"
    elif isinstance(data[0], np.int16):
        var_dtype = "int16_t"
    elif isinstance(data[0], np.integer):
        var_dtype = "int32_t"
    else:
        raise ValueError("Unsupported dtype")

    with open(dst_path, "w", encoding="UTF-8") as wfp:
        if is_header:
            wfp.write(f"#ifndef __{var_name.upper()}_H{os.linesep}")
            wfp.write(f"#define __{var_name.upper()}_H{os.linesep}")

        wfp.write(f"#include <cstdint>{os.linesep}{os.linesep}")

        wfp.write(f"const {var_dtype} {var_name}[] = {{{os.linesep}")
        for row in range(0, len(data), row_len):
            wfp.write("  " + ", ".join((str(val) for val in data[row : row + row_len])) + f", {os.linesep}")
        # END FOR
        wfp.write(f"}};{os.linesep}")
        wfp.write(f"const unsigned int {var_name}_len = {len(data)};{os.linesep}")
        if is_header:
            wfp.write(f"#endif // __{var_name.upper()}_H{os.linesep}")
    # END WITH


def xxd_c_dump(
    src_path: os.PathLike,
    dst_path: os.PathLike,
    var_name: str = "model",
    chunk_len: int = 12,
    is_header: bool = False,
):
    """Generate C like char array of hex values from binary source. Equivalent to `xxd -i src_path > dst_path`
        but with added features to provide # columns and variable name.

    Args:
        src_path (PathLike): Binary file source path
        dst_path (PathLike): C file destination path
        var_name (str, optional): C variable name. Defaults to 'model'.
        chunk_len (int, optional): Number of elements per row. Defaults to 12.
        is_header (bool): Write as header or source C file. Defaults to source.

    Example:

    ```python
    import helia_edge as helia

    src_path = "model.tflite"
    dst_path = "model.h"
    helia.converters.cpp.xxd_c_dump(
        src_path,
        dst_path,
        var_name="model",
        chunk_len=12,
        is_header=True
    )
    ```
    """
    var_len = 0
    with open(src_path, "rb", encoding=None) as rfp, open(dst_path, "w", encoding="UTF-8") as wfp:
        if is_header:
            wfp.write(f"#ifndef __{var_name.upper()}_H{os.linesep}")
            wfp.write(f"#define __{var_name.upper()}_H{os.linesep}")

        wfp.write(f"const unsigned char {var_name}[] = {{{os.linesep}")
        for chunk in iter(lambda: rfp.read(chunk_len), b""):
            wfp.write("  " + ", ".join((f"0x{c:02x}" for c in chunk)) + f", {os.linesep}")
            var_len += len(chunk)
        # END FOR
        wfp.write(f"}};{os.linesep}")
        wfp.write(f"const unsigned int {var_name}_len = {var_len};{os.linesep}")
        if is_header:
            wfp.write(f"#endif // __{var_name.upper()}_H{os.linesep}")
    # END WITH



"""
# Rescaling Layer API

This module provides classes to build rescaling layers.

Classes:
    Rescaling1D: Rescaling 1D
    Rescaling2D: Rescaling 2D

"""

import keras
from .base_augmentation import BaseAugmentation1D, BaseAugmentation2D
from ...utils import helia_export


@helia_export(path="helia_edge.layers.preprocessing.Rescaling1D")
class Rescaling1D(BaseAugmentation1D):
    scale: float

    def __init__(self, scale: float, **kwargs):
        """Rescale the input samples.

        Args:
            scale (float): The scaling factor.
        """
        super().__init__(**kwargs)
        self.scale = scale

    def augment_samples(self, inputs) -> keras.KerasTensor:
        """Rescale a batch of samples during training."""
        samples = inputs[self.SAMPLES]
        return samples * self.scale

    def compute_output_shape(self, input_shape, *args, **kwargs):
        """Compute output shape."""
        return input_shape

    def get_config(self):
        """Serialize the configuration."""
        config = super().get_config()
        config.update(scale=self.scale)
        return config


@helia_export(path="helia_edge.layers.preprocessing.Rescaling2D")
class Rescaling2D(BaseAugmentation2D):
    scale: float

    def __init__(self, scale: float, **kwargs):
        """Rescale the input samples.

        Args:
            scale (float): The scaling factor.
        """
        super().__init__(**kwargs)
        self.scale = scale

    def augment_samples(self, inputs) -> keras.KerasTensor:
        """Rescale a batch of samples during training."""
        samples = inputs[self.SAMPLES]
        return samples * self.scale

    def compute_output_shape(self, input_shape, *args, **kwargs):
        """Compute output shape."""
        return input_shape

    def get_config(self):
        """Serialize the configuration."""
        config = super().get_config()
        config.update(scale=self.scale)
        return config



"""
# Amplitude Warp Layer

Classes:
    AmplitudeWarp: Amplitude warping layer
"""

import keras

from .base_augmentation import BaseAugmentation1D
from ...utils import parse_factor, helia_export


@helia_export(path="helia_edge.layers.preprocessing.AmplitudeWarp")
class AmplitudeWarp(BaseAugmentation1D):
    sample_rate: float
    frequency: tuple[float, float]
    amplitude: tuple[float, float]
    noise_type: str

    def __init__(
        self,
        sample_rate: float = 1,
        frequency: float | tuple[float, float] = 100,
        amplitude: float | tuple[float, float] = 0.1,
        **kwargs,
    ):
        """Apply amplitude warping to the 1D input.
        Time points are first generated at given frequency resolution with amplitude picked from uniform distribution.
        These points are then interpolated to match the input duration and multiplied to the input.

        Args:
            sample_rate (float): Sample rate of the input.
            frequency (float|tuple[float,float]): Frequency of the warping in Hz. If tuple, frequency is randomly picked between the values.
            amplitude (float|tuple[float,float]): Amplitude of the warping. If tuple, amplitude is randomly picked between the values.

        Example:

        ```python
        sample_rate = 100 # Hz
        duration = 3*sample_rate # 3 seconds
        sig_freq = 10 # Hz
        sig_amp = 1 # Signal amplitude
        noise_freq = (1, 2) # Noise frequency range
        amplitude = (0.5, 2) # Noise amplitude range
        x = sig_amp*np.sin(2*np.pi*sig_freq*np.arange(duration)/sample_rate).reshape(-1, 1).astype(np.float32)
        x = keras.ops.convert_to_tensor(x)
        import helia_edge as helia

        lyr = helia.layers.preprocessing.AmplitudeWarp(
            sample_rate=sample_rate,
            frequency=noise_freq,
            amplitude=amplitude,
        )
        y = lyr(x)
        plt.plot(x.numpy())
        plt.plot(y.numpy())
        plt.show()
        ```
        """

        super().__init__(**kwargs)

        self.sample_rate = sample_rate
        self.frequency = parse_factor(frequency, min_value=None, max_value=sample_rate / 2, param_name="frequency")
        self.amplitude = parse_factor(amplitude, min_value=0, max_value=None, param_name="amplitude")

    def get_random_transformations(self, input_shape: tuple[int, int, int]) -> dict:
        """Generate noise distortion tensor

        Args:
            input_shape (tuple[int, ...]): Input shape.

        Returns:
            dict: Dictionary containing the noise tensor.
        """
        batch_size = input_shape[0]
        duration_size = input_shape[self.data_axis]
        ch_size = input_shape[self.ch_axis]

        # Add one period to the noise and clip later
        if self.frequency[0] == self.frequency[1]:
            frequency = keras.cast(self.frequency[0], dtype=self.compute_dtype)
        else:
            frequency = keras.random.uniform(
                shape=(),
                minval=self.frequency[0],
                maxval=self.frequency[1],
                seed=self.random_generator,
                dtype=self.compute_dtype,
            )

        duration_sec = keras.ops.cast(duration_size / self.sample_rate, dtype=self.compute_dtype)
        warp_duration = keras.ops.cast(duration_sec * frequency + frequency, dtype="int32")

        if self.data_format == "channels_first":
            warp_shape = (batch_size, 1, ch_size, warp_duration)
        else:
            warp_shape = (batch_size, 1, warp_duration, ch_size)

        warp_pts = keras.random.uniform(
            warp_shape, minval=self.amplitude[0], maxval=self.amplitude[1], seed=self.random_generator
        )

        # keras.ops doesnt contain any low-level interpolate. So we leverage the
        # image module and fix height to 1 as workaround
        warp = keras.ops.image.resize(
            warp_pts,
            size=(1, duration_size),
            interpolation="bicubic",
            crop_to_aspect_ratio=False,
            data_format=self.data_format,
        )
        # Remove height dimension
        warp = keras.ops.squeeze(warp, axis=1)
        return {"warp": warp}

    def augment_samples(self, inputs) -> keras.KerasTensor:
        """Augment all samples in the batch as it's faster."""
        samples = inputs[self.SAMPLES]
        if self.training:
            warp = inputs[self.TRANSFORMS]["warp"]
            return samples * warp
        return samples

    def get_config(self):
        """Serialize the layer configuration to a JSON-compatible dictionary."""
        config = super().get_config()
        config.update(
            {
                "sample_rate": self.sample,
                "frequency": self.frequency,
                "amplitude": self.amplitude,
                "noise_type": self.noise_type,
            }
        )
        return config



"""
# F-Score Metrics API

This module contains additional metrics for F-Score.

Classes:
    MultiF1Score: A wrapper around keras.metrics.F1Score to handle multi-dimensional data.

"""

import keras

from ..utils import helia_export


@helia_export(path="helia_edge.metrics.MultiF1Score")
class MultiF1Score(keras.metrics.F1Score):
    """A wrapper around keras.metrics.F1Score to handle multi-dimensional data.
    This class collapses down to 2D and treats last dimension as classes.

    Example:

    ```python
    import helia_edge as helia

    f1 = helia.metrics.MultiF1Score(average='macro')
    y_true = np.array([[0, 1, 0], [1, 0, 1]])
    y_pred = np.array([[0, 1, 0], [1, 0, 0]])
    f1.update_state(y_true, y_pred)
    print(f1.result().numpy())
    ```
    """

    def update_state(self, y_true, y_pred, sample_weight=None):
        y_true = keras.ops.convert_to_tensor(y_true, dtype=self.dtype)
        y_pred = keras.ops.convert_to_tensor(y_pred, dtype=self.dtype)
        # We force data to be 2D
        if len(y_true.shape) > 2:
            y_true = keras.ops.reshape(y_true, (-1, y_true.shape[-1]))
            y_pred = keras.ops.reshape(y_pred, (-1, y_pred.shape[-1]))

        super().update_state(y_true, y_pred, sample_weight)



"""
# Utils API

The `utils` module provides utility functions to help with common tasks such as downloading files, setting random seeds, and exporting data.

## Available Utilities

- **[AWS](aws.md)**: Provides functions to download files from AWS S3.
- **[Environment](env.md)**: Provides functions to set up the environment and logging.
- **[Export](export.md)**: Provides functions to export data.
- **[Factory](factory.md)**: Provides functions to create objects.
- **[File](file.md)**: Provides functions to download and save files.
- **[Preprocessing](preprocessing.md)**: Provides functions to preprocess data.
- **[RNG](rng.md)**: Provides functions to set random seeds.
- **[Tensor](tensor.md)**: Provides functions to work with tensors.

"""

from .aws import download_s3_file, download_s3_object, download_s3_objects
from .env import env_flag, setup_logger, silence_tensorflow
from .export import helia_export
from .factory import ItemFactory, create_factory
from .file import download_file, load_pkl, save_pkl, compute_checksum, resolve_template_path
from .preprocessing import (
    parse_factor,
    convert_inputs_to_tf_dataset,
    create_interleaved_dataset_from_generator,
    create_dataset_from_data,
    get_output_signature,
    get_output_signature_from_fn,
    get_output_signature_from_gen,
)
from .rng import set_random_seed, uniform_id_generator, random_id_generator
from .tensor import matches_spec



"""
# Frequency Mix Style Layer API

This module provides classes to perform frequency mix style augmentation.

Classes:
    FrequencyMixStyle2D: 2D frequency mix style augmentation

"""

import keras

from .base_augmentation import BaseAugmentation2D
from ...utils import parse_factor, helia_export


@helia_export(path="helia_edge.layers.preprocessing.FrequencyMixStyle2D")
class FrequencyMixStyle2D(BaseAugmentation2D):
    probability: float
    alpha: float
    epsilon: float

    def __init__(
        self,
        probability: float = 0.5,
        alpha: float = 1.0,
        epsilon: float = 1e-6,
        **kwargs,
    ):
        """Apply frequency mix style augmentation to the 2D input.

        Args:
            probability (float): Probability of applying the augmentation.
            alpha (float): Mixup alpha value.
            epsilon (float): Epsilon value for numerical stability.

        Example:

        ```python
            x = np.random.rand(4, 4, 3)
            lyr = FrequencyMixStyle2D(probability=1.0, alpha=1.0)
            y = lyr(x, training=True)
        ```
        """

        super().__init__(**kwargs)
        self.probability, _ = parse_factor([None, probability], min_value=0.0, max_value=1.0, param_name="probability")
        self.alpha, _ = parse_factor([None, alpha], min_value=0.0, max_value=None, param_name="alpha")
        self.epsilon, _ = parse_factor([None, epsilon], min_value=0.0, max_value=None, param_name="epsilon")

    def get_random_transformations(self, input_shape: tuple[int, int, int]) -> dict:
        """Generate noise distortion tensor

        Args:
            input_shape (tuple[int, ...]): Input shape.

        Returns:
            dict: Dictionary containing the noise tensor.
        """
        batch_size = input_shape[0]
        skip_augment = keras.random.uniform(
            shape=(), minval=0.0, maxval=1.0, dtype="float32", seed=self.random_generator
        )
        lmda = keras.random.beta(
            shape=(batch_size, 1, 1, 1), alpha=self.alpha, beta=self.alpha, seed=self.random_generator
        )
        perm = keras.random.shuffle(keras.ops.arange(batch_size), seed=self.random_generator)
        return {"lmda": lmda, "perm": perm, "skip_augment": skip_augment}

    def apply_mixstyle(self, x, lmda, perm):
        """Apply mixstyle augmentation

        Args:
            x (tf.Tensor): Input tensor
            lmda (tf.Tensor): Lambda tensor
            perm (tf.Tensor): Permutation tensor

        Returns:
            tf.Tensor: Augmented tensor
        """
        f_mu = keras.ops.mean(x, axis=[2, 3], keepdims=True)
        f_var = keras.ops.var(x, axis=[2, 3], keepdims=True)
        f_sig = keras.ops.sqrt(f_var + self.epsilon)

        x_normed = (x - f_mu) / f_sig
        f_mu_perm = keras.ops.take(f_mu, perm, axis=0)
        f_sig_perm = keras.ops.take(f_sig, perm, axis=0)
        x_perm = keras.ops.take(x_normed, perm, axis=0)
        x_mix = lmda * x_normed + (1 - lmda) * x_perm
        x_mix = x_mix * f_sig_perm + f_mu_perm
        return x_mix

    def augment_samples(self, inputs) -> keras.KerasTensor:
        """Augment samples

        Args:
            inputs (tf.Tensor): Input tensor

        Returns:
            tf.Tensor: Augmented tensor
        """

        samples = inputs[self.SAMPLES]
        transforms = inputs[self.TRANSFORMS]
        skip_augment = transforms["skip_augment"]
        if self.training:
            lmda = transforms["lmda"]
            perm = transforms["perm"]
            return keras.ops.cond(
                skip_augment > self.probability, lambda: samples, lambda: self.apply_mixstyle(samples, lmda, perm)
            )
        return samples


# def mixstyle(x, p=0.4, alpha=0.3, eps=1e-6):
#     if keras.random.uniform(shape=()) > p:
#         return x
#     batch_size = x.shape[0]

#     # x axis is NxFxTxC (batch_size, frequency, time, channels)

#     f_mu = keras.ops.mean(x, axis=[2, 3], keepdims=True)
#     f_var = keras.ops.var(x, axis=[2, 3], keepdims=True)
#     f_sig = keras.ops.sqrt(f_var + eps)
#     x_normed = (x - f_mu) / f_sig  # normalize input
#     lmda = keras.random.beta((batch_size, 1, 1, 1), alpha, alpha)
#     perm = keras.random.shuffle(keras.ops.arange(batch_size))
#     f_mu_perm = keras.ops.take(f_mu, perm, axis=0)
#     f_sig_perm = keras.ops.take(f_sig, perm, axis=0)

#     mu_mix = f_mu * lmda + f_mu_perm * (1 - lmda)  # generate mixed mean
#     sig_mix = f_sig * lmda + f_sig_perm * (1 - lmda)  # generate mixed standard deviation

#     x = x_normed * sig_mix + mu_mix  # denormalize input using the mixed frequency statistics
#     return x



"""
# File Utility API

This module provides utility functions to interact with files.

Functions:
    download_file: Download file from supplied url to destination streaming
    compute_checksum: Compute checksum of file
    load_pkl: Load pickled file
    save_pkl: Save python objects into pickle file
    resolve_template_path: Resolve templated path w/ supplied substitutions
"""

import gzip
import os
import hashlib
import pickle
from pathlib import Path
from string import Template
from typing import Any

import requests
from tqdm import tqdm


def download_file(src: str, dst: os.PathLike, progress: bool = True, chunk_size: int = 8192):
    """Download file from supplied url to destination streaming.

    Args:
        src (str): Source URL path
        dst (PathLike): Destination file path
        progress (bool, optional): Display progress bar. Defaults to True.

    """
    with requests.get(src, stream=True, timeout=3600 * 24) as r:
        r.raise_for_status()
        req_len = int(r.headers.get("Content-length", 0))
        prog_bar = tqdm(total=req_len, unit="iB", unit_scale=True) if progress else None
        with open(dst, "wb") as f:
            for chunk in r.iter_content(chunk_size=chunk_size):
                f.write(chunk)
                if prog_bar:
                    prog_bar.update(len(chunk))
            # END FOR
        # END WITH
    # END WITH


def compute_checksum(file: Path, checksum_type: str = "md5", chunk_size: int = 8192) -> str:
    """Compute checksum of file.

    Args:
        file (Path): File path
        checksum_type (str, optional): Checksum type. Defaults to "md5".
        chunk_size (int, optional): Chunk size. Defaults to 8192.

    Returns:
        str: Checksum value
    """
    if not file.is_file():
        raise FileNotFoundError(f"File {file} not found.")
    hash_algo = hashlib.new(checksum_type)
    with open(file, "rb") as f:
        for chunk in iter(lambda: f.read(chunk_size), b""):
            hash_algo.update(chunk)
    return hash_algo.hexdigest()


def load_pkl(file: str, compress: bool = True) -> dict[str, Any]:
    """Load pickled file.

    Args:
        file (str): File path (.pkl)
        compress (bool, optional): If file is compressed. Defaults to True.

    Returns:
        dict[str, Any]: Dictionary of pickled objects
    """
    if compress:
        with gzip.open(file, "rb") as fh:
            return pickle.load(fh)
    else:
        with open(file, "rb") as fh:
            return pickle.load(fh)


def save_pkl(file: str, compress: bool = True, **kwargs):
    """Save python objects into pickle file.

    Args:
        file (str): File path (.pkl)
        compress (bool, optional): Whether to compress file. Defaults to True.
    """
    if compress:
        with gzip.open(file, "wb") as fh:
            pickle.dump(kwargs, fh, protocol=4)
    else:
        with open(file, "wb") as fh:
            pickle.dump(kwargs, fh, protocol=4)


def resolve_template_path(fpath: Path, **kwargs: Any) -> Path:
    """Resolve templated path w/ supplied substitutions.

    Args:
        fpath (Path): File path
        **kwargs (Any): Template arguments

    Returns:
        Path: Resolved file path
    """
    return Path(Template(str(fpath)).safe_substitute(**kwargs))



"""
# SimCLR Loss

This module implements the SimCLR loss function for contrastive self-supervised learning.

Classes:
    SimCLRLoss: Implements SimCLR Cosine Similarity loss.

Functions:
    l2_normalize: Normalizes a tensor along a given axis.

"""

import keras

from ..utils import helia_export

LARGE_NUM = 1e9


def l2_normalize(x: keras.KerasTensor, axis: int | tuple[int, ...] | None = None) -> keras.KerasTensor:
    """Performs L2 normalization on a tensor along a given axis.

    Args:
        x (tf.Tensor): Input tensor
        axis (int | tuple[int], optional): Axis. Defaults to None.

    Returns:
        tf.Tensor: Normalized tensor
    """
    epsilon = keras.backend.epsilon()
    power_sum = keras.ops.sum(keras.ops.square(x), axis=axis, keepdims=True)
    norm = keras.ops.reciprocal(keras.ops.sqrt(keras.ops.maximum(power_sum, epsilon)))
    return keras.ops.multiply(x, norm)


@helia_export(path="helia_edge.losses.SimCLRLoss")
class SimCLRLoss(keras.losses.Loss):
    """Implements SimCLR Cosine Similarity loss.

    SimCLR loss is used for contrastive self-supervised learning.

    Args:
        temperature (float): A scaling factor for cosine similarity b/w [0, 1].

    References:
        - [SimCLR paper](https://arxiv.org/pdf/2002.05709)
    """

    def __init__(self, temperature: float, **kwargs):
        super().__init__(**kwargs)
        self.temperature = temperature

    def call(self, projections_1: keras.KerasTensor, projections_2: keras.KerasTensor) -> keras.KerasTensor:
        """Computes SimCLR loss for a pair of projections in a contrastive
        learning trainer.

        Note that unlike most loss functions, this should not be called with
        y_true and y_pred, but with two unlabeled projections. It can otherwise
        be treated as a normal loss function.

        Args:
            projections_1 (keras.KerasTensor): a tensor with the output of the first projection
                model in a contrastive learning trainer
            projections_2 (keras.KerasTensor): a tensor with the output of the second projection
                model in a contrastive learning trainer

        Returns:
            keras.KerasTensor: A tensor with the SimCLR loss computed from the input projections
        """
        # Normalize the projections
        projections_1 = l2_normalize(projections_1, axis=1)
        projections_2 = l2_normalize(projections_2, axis=1)

        # Produce artificial labels, 1 for each image in the batch.
        batch_size = keras.ops.shape(projections_1)[0]
        labels = keras.ops.one_hot(keras.ops.arange(batch_size), batch_size * 2)
        masks = keras.ops.one_hot(keras.ops.arange(batch_size), batch_size)

        # Compute logits
        logits_11 = keras.ops.matmul(projections_1, keras.ops.transpose(projections_1)) / self.temperature
        logits_11 = logits_11 - keras.ops.cast(masks * LARGE_NUM, logits_11.dtype)
        logits_22 = keras.ops.matmul(projections_2, keras.ops.transpose(projections_2)) / self.temperature
        logits_22 = logits_22 - keras.ops.cast(masks * LARGE_NUM, logits_22.dtype)
        logits_12 = keras.ops.matmul(projections_1, keras.ops.transpose(projections_2)) / self.temperature
        logits_21 = keras.ops.matmul(projections_2, keras.ops.transpose(projections_1)) / self.temperature

        loss_a = keras.losses.categorical_crossentropy(
            labels, keras.ops.concatenate([logits_12, logits_11], 1), from_logits=True
        )
        loss_b = keras.losses.categorical_crossentropy(
            labels, keras.ops.concatenate([logits_21, logits_22], 1), from_logits=True
        )

        return loss_a + loss_b

    def get_config(self):
        config = super().get_config()
        config.update({"temperature": self.temperature})
        return config



"""Convolutional Layers API

This module provides classes to build convolutional layers.

Functions:
    conv1d: 1D convolutional layer using 2D convolutional layer
    conv2d: 2D convolutional layer

Please refer to [Keras Convolutional Layers](https://keras.io/api/layers/convolution_layers/) for additional layers.

"""

import keras


def conv2d(
    filters: int,
    kernel_size: int | tuple[int, int] = 3,
    strides: int | tuple[int, int] = 1,
    padding: str = "same",
    use_bias: bool = False,
    groups: int = 1,
    dilation: int = 1,
    name: str | None = None,
) -> keras.Layer:
    """2D convolutional layer

    Args:
        filters (int): Number of filters
        kernel_size (int | tuple[int, int], optional): Kernel size. Defaults to 3.
        strides (int | tuple[int, int], optional): Stride length. Defaults to 1.
        padding (str, optional): Padding. Defaults to "same".
        use_bias (bool, optional): Add bias. Defaults to False.
        groups (int, optional): Number of groups. Defaults to 1.
        dilation (int, optional): Dilation rate. Defaults to 1.
        name (str | None, optional): Layer name. Defaults to None.

    Returns:
        keras.Layer: Functional 2D conv layer
    """
    name = name + ".conv" if name else None
    return keras.layers.Conv2D(
        filters,
        kernel_size=kernel_size,
        strides=strides,
        padding=padding,
        use_bias=use_bias,
        groups=groups,
        dilation_rate=dilation,
        kernel_initializer="he_normal",
        name=name,
    )


def conv1d(
    filters: int,
    kernel_size: int = 3,
    strides: int = 1,
    padding: str = "same",
    use_bias: bool = False,
    name: str | None = None,
) -> keras.Layer:
    """1D convolutional layer using 2D convolutional layer

    Args:
        filters (int): Number of filters
        kernel_size (int, optional): Kernel size. Defaults to 3.
        strides (int, optional): Stride length. Defaults to 1.
        padding (str, optional): Padding. Defaults to "same".
        use_bias (bool, optional): Add bias. Defaults to False.
        name (str | None, optional): Layer name. Defaults to None.

    Returns:
        keras.Layer: Functional 1D conv layer
    """
    name = name + ".conv" if name else None
    return keras.layers.Conv2D(
        filters,
        kernel_size=(1, kernel_size),
        strides=(1, strides),
        padding=padding,
        use_bias=use_bias,
        kernel_initializer="he_normal",
        name=name,
    )



"""
# Random Crop Layers API

This module provides classes to build random crop layers.

Classes:
    RandomCrop1D: Random crop 1D
    RandomCrop2D: Random crop 2D

"""

import keras
from .base_augmentation import BaseAugmentation1D, BaseAugmentation2D
from ...utils import helia_export


@helia_export(path="helia_edge.layers.preprocessing.RandomCrop1D")
class RandomCrop1D(BaseAugmentation1D):
    duration: int
    unique_batch: bool

    def __init__(self, duration: int, unique_batch: bool = False, **kwargs):
        """Randomly crop 1D input samples.

        Args:
            duration (int): Duration of the output samples.
            unique_batch (bool): If True, each sample in the batch will have a unique crop.

        Example:

        ```python
            duration = 100
            lyr = RandomCrop1D(duration=duration)
            x = np.random.randn(32, 1000, 1)
            y = lyr(x, training=True)
        ```
        """

        super().__init__(**kwargs)
        self.duration = duration
        self.unique_batch = unique_batch

    def random_crop(self, sample, start):
        """Randomly crop single sample"""
        ch_size = keras.ops.shape(sample)[self.ch_axis]
        if self.data_format == "channels_first":
            return keras.ops.slice(sample, [0, start], [ch_size, self.duration])
        return keras.ops.slice(sample, [start, 0], [self.duration, ch_size])
        # END IF

    # END DEF

    def get_random_transformations(self, input_shape):
        """Generate random start indices for cropping."""
        batch_size = input_shape[0]
        duration_size = input_shape[self.data_axis]
        if duration_size < self.duration:
            raise ValueError(f"Input duration ({duration_size}) must be greater than output duration ({self.duration})")

        d_diff = duration_size - self.duration
        if self.unique_batch:
            start = keras.random.randint(
                shape=(batch_size,), minval=0, maxval=int(d_diff + 1), seed=self.random_generator, dtype="int32"
            )
        else:
            start = keras.random.randint(
                shape=(), minval=0, maxval=int(d_diff + 1), seed=self.random_generator, dtype="int32"
            )
            start = keras.ops.broadcast_to(start, [batch_size])
        return {"start": start}

    def augment_sample(self, inputs) -> keras.KerasTensor:
        """Augment single sample with random crop."""
        sample = inputs[self.SAMPLES]
        start = inputs[self.TRANSFORMS]["start"]

        if self.training:
            sample = self.random_crop(sample, start)
        return sample

    def compute_output_shape(self, input_shape, *args, **kwargs):
        """Compute the output shape of the layer."""
        output_shape = list(input_shape)
        output_shape[self.data_axis] = self.duration
        return tuple(input_shape)

    def get_config(self):
        """Serializes the configuration of the layer."""
        config = super().get_config()
        config.update(
            {
                "duration": self.duration,
                "unique_batch": self.unique_batch,
            }
        )
        return config


@helia_export(path="helia_edge.layers.preprocessing.RandomCrop2D")
class RandomCrop2D(BaseAugmentation2D):
    height: int
    width: int
    unique_batch: bool

    def __init__(self, height: int, width: int, unique_batch: bool = False, **kwargs):
        """Randomly crop 2D input samples.

        Args:
            height (int): Height of the output samples.
            width (int): Width of the output samples.
            unique_batch (bool): If True, each sample in the batch will have a unique crop.

        Example:

        ```python
            height = 32
            width = 32
            lyr = RandomCrop2D(height=height, width=width)
            x = np.random.randn(32, 64, 64, 3)
            y = lyr(x, training=True)
        ```
        """

        super().__init__(**kwargs)
        self.height = height
        self.width = width
        self.unique_batch = unique_batch

    def random_crop(self, sample, start_h, start_w):
        """Randomly crop single sample"""
        if self.data_format == "channels_first":
            return keras.ops.slice(sample, [0, start_h, start_w], [-1, self.height, self.width])
        return keras.ops.slice(sample, [start_h, start_w, 0], [self.height, self.width, -1])
        # END IF

    def get_random_transformations(self, input_shape):
        """Generate random start indices for cropping."""
        batch_size = input_shape[0]
        height_size = input_shape[self.height_axis]
        width_size = input_shape[self.width_axis]
        if height_size < self.height:
            raise ValueError(f"Input height ({height_size}) must be greater than output height ({self.height})")
        if width_size < self.width:
            raise ValueError(f"Input width ({width_size}) must be greater than output width ({self.width})")

        h_diff = height_size - self.height
        w_diff = width_size - self.width
        if self.unique_batch:
            start_h = keras.random.randint(
                shape=(batch_size,), minval=0, maxval=int(h_diff + 1), seed=self.random_generator, dtype="int32"
            )
            start_w = keras.random.randint(
                shape=(batch_size,), minval=0, maxval=int(w_diff + 1), seed=self.random_generator, dtype="int32"
            )
        else:
            start_h = keras.random.randint(
                shape=(), minval=0, maxval=int(h_diff + 1), seed=self.random_generator, dtype="int32"
            )
            start_h = keras.ops.broadcast_to(start_h, [batch_size])
            start_w = keras.random.randint(
                shape=(), minval=0, maxval=int(w_diff + 1), seed=self.random_generator, dtype="int32"
            )
            start_w = keras.ops.broadcast_to(start_w, [batch_size])
        return {"start_h": start_h, "start_w": start_w}

    def augment_sample(self, inputs) -> keras.KerasTensor:
        """Augment single sample with random crop."""
        sample = inputs[self.SAMPLES]
        start_h = inputs[self.TRANSFORMS]["start_h"]
        start_w = inputs[self.TRANSFORMS]["start_w"]

        if self.training:
            sample = self.random_crop(sample, start_h, start_w)
        return sample

    def compute_output_shape(self, input_shape, *args, **kwargs):
        """Compute the output shape of the layer."""
        output_shape = list(input_shape)
        output_shape[self.height_axis] = self.height
        output_shape[self.width_axis] = self.width
        return tuple(input_shape)

    def get_config(self):
        """Serializes the configuration of the layer."""
        config = super().get_config()
        config.update(
            {
                "height": self.height,
                "width": self.width,
                "unique_batch": self.unique_batch,
            }
        )
        return config



"""
# Random Choice Layer API

This module provides classes to build random choice layers.

Classes:
    RandomChoice: Random choice

"""

import keras
from .base_augmentation import BaseAugmentation
from ...utils import helia_export


@helia_export(path="helia_edge.layers.preprocessing.RandomChoice")
class RandomChoice(BaseAugmentation):
    layers: list[BaseAugmentation]
    batchwise: bool

    def __init__(self, layers: list[BaseAugmentation], batchwise: bool = False, **kwargs):
        """Randomly choose one augmentation layer from a list of augmentation layers.

        Args:
            layers (list[BaseAugmentation]): List of augmentation layers to choose from.
            batchwise (bool): If True, apply same layer to all samples in the batch.
        """
        super().__init__(**kwargs)
        self.layers = layers
        self.batchwise = batchwise

    def batch_augment(self, inputs):
        """Apply random layer(s) to the batch"""
        # If batchwise, apply the same layer to all samples in the batch
        if self.batchwise:
            lyr_idx: int = keras.random.randint(
                shape=(), minval=0, maxval=len(self.layers), dtype="int32", seed=self.random_generator
            )
            branch_fns = [lambda x: layer.call(x, training=self.training) for layer in self.layers]
            return keras.ops.switch(lyr_idx, branch_fns, inputs)
        raise NotImplementedError("Batchwise=False is not implemented yet")

    def get_config(self):
        """Serializes the configuration of the layer."""
        config = super().get_config()
        config.update(
            {
                "layers": self.layers,
                "batchwise": self.batchwise,
            }
        )
        return config



"""
# U-NeXt

## Overview

U-NeXt is a modification of U-Net that utilizes techniques from ResNeXt and EfficientNetV2. During the encoding phase, mbconv blocks are used to efficiently process the input.

Classes:
    UNextParams: U-NeXt parameters
    UNextModel: Helper class to generate

Functions:
    unext_block: Create U-NeXt block
    se_block: Squeeze and excite block
    norm_layer: Normalization layer
    unext_core: Create U-NeXt core
    unext_layer: Create U-NeXt layer

## Additions

The U-NeXt architecture has been modified to allow the following:

* MBConv blocks used in the encoding phase.
* Squeeze and excitation (SE) blocks added within blocks.

"""

from typing import Literal

import keras
from pydantic import BaseModel, Field


class UNextBlockParams(BaseModel):
    """UNext block parameters

    Attributes:
        filters (int): Number of filters
        depth (int): Layer depth
        ddepth (int | None): Layer decoder depth
        kernel (int | tuple[int, int]): Kernel size
        pool (int | tuple[int, int]): Pool size
        strides (int | tuple[int, int]): Stride size
        skip (bool): Add skip connection
        expand_ratio (float): Expansion ratio
        se_ratio (float): Squeeze and excite ratio
        dropout (float | None): Dropout rate
        norm (Literal["batch", "layer"] | None): Normalization type

    """

    filters: int = Field(..., description="# filters")
    depth: int = Field(default=1, description="Layer depth")
    ddepth: int | None = Field(default=None, description="Layer decoder depth")
    kernel: int | tuple[int, int] = Field(default=3, description="Kernel size")
    pool: int | tuple[int, int] = Field(default=2, description="Pool size")
    strides: int | tuple[int, int] = Field(default=2, description="Stride size")
    skip: bool = Field(default=True, description="Add skip connection")
    expand_ratio: float = Field(default=1, description="Expansion ratio")
    se_ratio: float = Field(default=0, description="Squeeze and excite ratio")
    dropout: float | None = Field(default=None, description="Dropout rate")
    norm: Literal["batch", "layer"] | None = Field(default="layer", description="Normalization type")


class UNextParams(BaseModel):
    """UNext parameters

    Attributes:
        blocks (list[UNextBlockParams]): UNext blocks
        include_top (bool): Include top
        use_logits (bool): Use logits
        output_kernel_size (int | tuple[int, int]): Output kernel size
        output_kernel_stride (int | tuple[int, int]): Output kernel stride
        name (str): Model name

    """

    blocks: list[UNextBlockParams] = Field(default_factory=list, description="UNext blocks")
    include_top: bool = Field(default=True, description="Include top")
    use_logits: bool = Field(default=True, description="Use logits")
    output_kernel_size: int | tuple[int, int] = Field(default=3, description="Output kernel size")
    output_kernel_stride: int | tuple[int, int] = Field(default=1, description="Output kernel stride")
    name: str = Field(default="UNext", description="Model name")


def se_block(ratio: int = 8, name: str | None = None):
    """Squeeze and excite block"""

    def layer(x: keras.KerasTensor) -> keras.KerasTensor:
        num_chan = x.shape[-1]
        # Squeeze
        y = keras.layers.GlobalAveragePooling2D(name=f"{name}_pool" if name else None, keepdims=True)(x)

        y = keras.layers.Conv2D(
            num_chan // ratio,
            kernel_size=1,
            use_bias=True,
            name=f"{name}_sq" if name else None,
        )(y)

        y = keras.layers.Activation("relu6", name=f"{name}_relu" if name else None)(y)

        # Excite
        y = keras.layers.Conv2D(num_chan, kernel_size=1, use_bias=True, name=f"{name}_ex" if name else None)(y)
        y = keras.layers.Activation(keras.activations.hard_sigmoid, name=f"{name}_sigg" if name else None)(y)
        y = keras.layers.Multiply(name=f"{name}_mul" if name else None)([x, y])
        return y

    return layer


def norm_layer(norm: str, name: str) -> keras.Layer:
    """Normalization layer

    Args:
        norm (str): Normalization type
        name (str): Name

    Returns:
        keras.Layer: Layer
    """

    def layer(x: keras.KerasTensor) -> keras.KerasTensor:
        """Functional normalization layer

        Args:
            x (keras.KerasTensor): Input tensor

        Returns:
            keras.KerasTensor: Output tensor
        """
        if norm == "batch":
            return keras.layers.BatchNormalization(axis=-1, name=f"{name}_BN")(x)
        if norm == "layer":
            ln_axis = 2 if x.shape[1] == 1 else 1 if x.shape[2] == 1 else (1, 2)
            return keras.layers.LayerNormalization(axis=ln_axis, name=f"{name}_LN")(x)
        return x

    return layer


def unext_block(
    output_filters: int,
    expand_ratio: float = 1,
    kernel_size: int | tuple[int, int] = 3,
    strides: int | tuple[int, int] = 1,
    se_ratio: float = 4,
    dropout: float | None = 0,
    norm: Literal["batch", "layer"] | None = "batch",
    name: str | None = None,
) -> keras.Layer:
    """Create UNext block"""

    def layer(x: keras.KerasTensor) -> keras.KerasTensor:
        input_filters: int = x.shape[-1]
        strides_len = strides if isinstance(strides, int) else sum(strides) // len(strides)
        add_residual = input_filters == output_filters and strides_len == 1
        ln_axis = 2 if x.shape[1] == 1 else 1 if x.shape[2] == 1 else (1, 2)

        # Depthwise conv
        y = keras.layers.Conv2D(
            input_filters,
            kernel_size=kernel_size,
            groups=input_filters,
            strides=1,
            padding="same",
            use_bias=norm is None,
            kernel_initializer="he_normal",
            kernel_regularizer=keras.regularizers.L2(1e-3),
            name=f"{name}_dwconv" if name else None,
        )(x)
        if norm == "batch":
            y = keras.layers.BatchNormalization(
                name=f"{name}_norm",
            )(y)
        elif norm == "layer":
            y = keras.layers.LayerNormalization(
                axis=ln_axis,
                name=f"{name}_norm" if name else None,
            )(y)
        # END IF

        # Inverted expansion block
        if expand_ratio != 1:
            y = keras.layers.Conv2D(
                filters=int(expand_ratio * input_filters),
                kernel_size=1,
                strides=1,
                padding="same",
                use_bias=norm is None,
                groups=input_filters,
                kernel_initializer="he_normal",
                kernel_regularizer=keras.regularizers.L2(1e-3),
                name=f"{name}_expand" if name else None,
            )(y)

            y = keras.layers.Activation(
                "relu6",
                name=f"{name}_relu" if name else None,
            )(y)

        # Squeeze and excite
        if se_ratio > 1:
            name_se = f"{name}_se" if name else None
            y = se_block(ratio=se_ratio, name=name_se)(y)

        y = keras.layers.Conv2D(
            filters=output_filters,
            kernel_size=1,
            strides=1,
            padding="same",
            use_bias=norm is None,
            kernel_initializer="he_normal",
            kernel_regularizer=keras.regularizers.L2(1e-3),
            name=f"{name}_project" if name else None,
        )(y)

        if add_residual:
            if dropout and dropout > 0:
                y = keras.layers.Dropout(
                    dropout,
                    noise_shape=(y.shape),
                    name=f"{name}_drop" if name else None,
                )(y)
            y = keras.layers.Add(name=f"{name}_res" if name else None)([x, y])
        return y

    # END DEF
    return layer


def unext_core(
    x: keras.KerasTensor,
    params: UNextParams,
) -> keras.KerasTensor:
    """Create UNext TF functional core

    Args:
        x (keras.KerasTensor): Input tensor
        params (UNextParams): Model parameters.

    Returns:
        keras.KerasTensor: Output tensor
    """

    y = x

    #### ENCODER ####
    skip_layers: list[keras.layers.Layer | None] = []
    for i, block in enumerate(params.blocks):
        name = f"ENC{i + 1}"
        for d in range(block.depth):
            y = unext_block(
                output_filters=block.filters,
                expand_ratio=block.expand_ratio,
                kernel_size=block.kernel,
                strides=1,
                se_ratio=block.se_ratio,
                dropout=block.dropout,
                norm=block.norm,
                name=f"{name}_D{d + 1}",
            )(y)
        # END FOR
        skip_layers.append(y if block.skip else None)

        # Downsample using strided conv
        y = keras.layers.Conv2D(
            filters=block.filters,
            kernel_size=block.pool,
            strides=block.strides,
            padding="same",
            use_bias=block.norm is None,
            kernel_initializer="he_normal",
            kernel_regularizer=keras.regularizers.L2(1e-3),
            name=f"{name}_pool",
        )(y)
        if block.norm == "batch":
            y = keras.layers.BatchNormalization(
                name=f"{name}_norm",
            )(y)
        elif block.norm == "layer":
            ln_axis = 2 if y.shape[1] == 1 else 1 if y.shape[2] == 1 else (1, 2)
            y = keras.layers.LayerNormalization(
                axis=ln_axis,
                name=f"{name}_norm",
            )(y)
        # END IF
    # END FOR

    #### DECODER ####
    for i, block in enumerate(reversed(params.blocks)):
        name = f"DEC{i + 1}"
        for d in range(block.ddepth or block.depth):
            y = unext_block(
                output_filters=block.filters,
                expand_ratio=block.expand_ratio,
                kernel_size=block.kernel,
                strides=1,
                se_ratio=block.se_ratio,
                dropout=block.dropout,
                norm=block.norm,
                name=f"{name}_D{d + 1}",
            )(y)
        # END FOR

        # Upsample using transposed conv
        # y = keras.layers.Conv1DTranspose(
        #     filters=block.filters,
        #     kernel_size=block.pool,
        #     strides=block.strides,
        #     padding="same",
        #     kernel_initializer="he_normal",
        #     kernel_regularizer=keras.regularizers.L2(1e-3),
        #     name=f"{name}_unpool",
        # )(y)

        y = keras.layers.Conv2D(
            filters=block.filters,
            kernel_size=block.pool,
            strides=1,
            padding="same",
            use_bias=block.norm is None,
            kernel_initializer="he_normal",
            kernel_regularizer=keras.regularizers.L2(1e-3),
            name=f"{name}_conv",
        )(y)
        y = keras.layers.UpSampling2D(size=block.strides, name=f"{name}_unpool")(y)

        # Skip connection
        skip_layer = skip_layers.pop()
        if skip_layer is not None:
            # y = keras.layers.Concatenate(name=f"{name}_S1_cat")([y, skip_layer])
            y = keras.layers.Add(name=f"{name}_S1_cat")([y, skip_layer])

            # Use conv to reduce filters
            y = keras.layers.Conv2D(
                block.filters,
                kernel_size=1,  # block.kernel,
                padding="same",
                kernel_initializer="he_normal",
                kernel_regularizer=keras.regularizers.L2(1e-3),
                use_bias=block.norm is None,
                name=f"{name}_S1_conv",
            )(y)

            if block.norm == "batch":
                y = keras.layers.BatchNormalization(
                    name=f"{name}_S1_norm",
                )(y)
            elif block.norm == "layer":
                ln_axis = 2 if y.shape[1] == 1 else 1 if y.shape[2] == 1 else (1, 2)
                y = keras.layers.LayerNormalization(
                    axis=ln_axis,
                    name=f"{name}_S1_norm",
                )(y)
            # END IF

            y = keras.layers.Activation(
                "relu6",
                name=f"{name}_S1_relu" if name else None,
            )(y)
        # END IF

        y = unext_block(
            output_filters=block.filters,
            expand_ratio=block.expand_ratio,
            kernel_size=block.kernel,
            strides=1,
            se_ratio=block.se_ratio,
            dropout=block.dropout,
            norm=block.norm,
            name=f"{name}_D{block.depth + 1}",
        )(y)

    # END FOR
    return y


def unext_layer(
    inputs: keras.KerasTensor,
    params: UNextParams,
    num_classes: int | None = None,
) -> keras.KerasTensor:
    """Create UNext TF functional model

    Args:
        inputs (keras.KerasTensor): Input tensor
        params (UNextParams): Model parameters.
        num_classes (int, optional): Number of classes.

    Returns:
        keras.KerasTensor: Output tensor
    """
    requires_reshape = len(inputs.shape) == 3
    if requires_reshape:
        y = keras.layers.Reshape((1,) + inputs.shape[1:])(inputs)
    else:
        y = inputs

    y = unext_core(y, params)

    if params.include_top:
        # Add a per-point classification layer
        y = keras.layers.Conv2D(
            num_classes,
            kernel_size=params.output_kernel_size,
            padding="same",
            kernel_initializer="he_normal",
            kernel_regularizer=keras.regularizers.L2(1e-3),
            name="NECK_conv",
            use_bias=True,
        )(y)
        if not params.use_logits:
            y = keras.layers.Softmax()(y)
        # END IF
    # END IF

    # Always reshape back to original shape
    if requires_reshape:
        y = keras.layers.Reshape(y.shape[2:])(y)

    return y


class UNextModel:
    """Helper class to generate model from parameters"""

    @staticmethod
    def layer_from_params(inputs: keras.Input, params: UNextParams | dict, num_classes: int | None = None):
        """Create layer from parameters"""
        if isinstance(params, dict):
            params = UNextParams(**params)
        return unext_layer(x=inputs, params=params, num_classes=num_classes)

    @staticmethod
    def model_from_params(inputs: keras.Input, params: UNextParams | dict, num_classes: int | None = None):
        """Create model from parameters"""
        outputs = UNextModel.layer_from_params(inputs=inputs, params=params, num_classes=num_classes)
        return keras.Model(inputs=inputs, outputs=outputs)



"""
# Random Channel Layer API

This module provides classes to build random channel layers.

Classes:
    RandomChannel: Random channel

"""

import keras

from .base_augmentation import BaseAugmentation
from ...utils import helia_export


@helia_export(path="helia_edge.layers.preprocessing.RandomChannel")
class RandomChannel(BaseAugmentation):
    batchwise: bool

    def __init__(self, batchwise: bool = False, **kwargs):
        """Randomly picks a single channel from the input samples.

        Args:
            batchwise (bool): If True, grabs same channel from all samples in batch.

        """
        super().__init__(**kwargs)
        self.batchwise = batchwise

    def get_random_transformations(self, input_shape):
        """Generate random channel index."""

        batch = input_shape[0]
        if self.data_format == "channels_first":
            ch_size = input_shape[self.ch_axis]
        else:
            ch_size = input_shape[self.ch_axis]

        # Pick same channel for all samples in the batch
        if self.batchwise:
            ch_idx = keras.random.randint(shape=(), minval=0, maxval=ch_size, seed=self.random_generator, dtype="int32")
            ch_idx = keras.ops.broadcast_to(ch_idx, [batch])
        else:
            ch_idx = keras.random.randint(
                shape=(batch,), minval=0, maxval=ch_size, seed=self.random_generator, dtype="int32"
            )
        return {"channel": ch_idx}

    def augment_samples(self, inputs):
        """Augment samples"""
        # If batchwise, easier to grab the channel here
        if self.batchwise:
            samples = inputs[self.SAMPLES]
            channel = inputs[self.TRANSFORMS]["channel"]
            channel = keras.ops.take(channel, 0, axis=0)
            return keras.ops.expand_dims(keras.ops.take(samples, channel, axis=self.ch_axis), axis=self.ch_axis)
        # Otherwise let the augment_sample method handle it
        else:
            super().augment_samples(inputs)

    def augment_sample(self, inputs) -> keras.KerasTensor:
        """Augment a sample during training."""
        sample = inputs[self.SAMPLES]
        channel = inputs[self.TRANSFORMS]["channel"]
        keras.ops.expand_dims(keras.ops.take(sample, channel, axis=self.ch_axis), axis=self.ch_axis)



"""
# RegNet Model API

This module provides utility functions to generate RegNet models.

Fore more information, please refer to the following paper: https://arxiv.org/abs/2101.00590

Classes:
    RegNetBlockParam: RegNet block parameters
    RegNetParams: RegNet parameters
    RegNetModel: Helper class to generate RegNet models

Functions:
    regnet_core: RegNet core
    regnet_layer: Generate RegNet model

"""

from typing import Callable, Literal

import keras
from pydantic import BaseModel, Field

from ..layers.normalization import batch_normalization
from ..layers.convolutional import conv2d
from ..layers.squeeze_excite import se_layer

from .utils import make_divisible


class RegNetBlockParam(BaseModel):
    """RegNet block parameters

    Attributes:
        filters (int): Number of filters
        depth (int): Layer depth
        group_width (int): Group width
        kernel_size (int | tuple[int, int]): Kernel size
        strides (int | tuple[int, int]): Stride size
        se_ratio (float): Squeeze Excite ratio
        droprate (float): Drop rate
        activation (str): Activation function

    """

    filters: int = Field(..., description="# filters")
    depth: int = Field(default=1, description="Layer depth")
    group_width: int = Field(default=1, description="Group width. Must be divisible by in/out filters")
    kernel_size: int | tuple[int, int] = Field(default=3, description="Kernel size")
    strides: int | tuple[int, int] = Field(default=1, description="Stride size")
    se_ratio: float = Field(default=8, description="Squeeze Excite ratio")
    droprate: float = Field(default=0, description="Drop rate")
    activation: str = Field(default="relu6", description="Activation function")


class RegNetParams(BaseModel):
    """RegNet parameters

    Attributes:

        blocks (list[RegNetBlockParam]): RegNet blocks
        input_filters (int): Input filters
        input_strides (int | tuple[int, int]): Input stride
        input_activation (str): Input activation
        output_filters (int): Output filters
        block_style (Literal["y", "z"]): Block style
        include_top (bool): Include top
        output_activation (str | None): Output activation
        dropout (float): Dropout rate
        name (str): Model name

    """

    blocks: list[RegNetBlockParam] = Field(default_factory=list, description="RegNet blocks")
    input_filters: int = Field(default=0, description="Input filters")
    input_strides: int | tuple[int, int] = Field(default=2, description="Input stride")
    input_activation: str = Field(default="relu6", description="Input activation")
    output_filters: int = Field(default=0, description="Output filters")
    block_style: Literal["y", "z"] = Field(default="y", description="Block style")
    include_top: bool = Field(default=True, description="Include top")
    output_activation: str | None = Field(default=None, description="Output activation")
    dropout: float = Field(default=0.2, description="Dropout rate")
    name: str = Field(default="RegNet", description="Model name")


def yblock(
    output_filters: int = 0,
    group_width: int = 0,
    kernel_size: int | tuple[int, int] = 3,
    strides: int | tuple[int, int] = 1,
    se_ratio: float = 4,
    activation: str = "relu6",
    name: str | None = None,
) -> Callable[[keras.KerasTensor], keras.KerasTensor]:
    """RegNet Y-Block

    Args:
        output_filters (int, optional): Number of output filters. Defaults to 0.
        group_width (int, optional): Group width. Defaults to 0.
        kernel_size (int | tuple[int, int], optional): Kernel size. Defaults to 3.
        strides (int | tuple[int, int], optional): Stride length. Defaults to 1.
        se_ratio (float, optional): SE ratio. Defaults to 4.
        activation (str, optional): Activation function. Defaults to "relu6".
        name (str|None, optional): Block name. Defaults to None.

    Returns:
        Callable[[keras.KerasTensor], keras.KerasTensor]: Functional layer
    """

    def layer(x: keras.KerasTensor) -> keras.KerasTensor:
        input_filters = x.shape[-1]
        groups = output_filters // group_width
        use_skip = (input_filters != output_filters) or (strides != 1 if isinstance(strides, int) else strides[0] != 1)

        name_ex = f"{name}_exp" if name else None
        y = conv2d(output_filters, kernel_size=(1, 1), name=name_ex)(x)
        y = batch_normalization(name=name_ex)(y)
        y = keras.layers.Activation(activation, name=f"{name_ex}_act")(y)

        y = keras.layers.Conv2D(
            output_filters,
            kernel_size=kernel_size,
            strides=strides,
            groups=groups,
            padding="same",
            use_bias=False,
            kernel_initializer=keras.initializers.VarianceScaling(),  # type: ignore
        )(y)
        y = batch_normalization()(y)
        y = keras.layers.Activation(activation)(y)

        if se_ratio:
            y = se_layer(ratio=se_ratio)(y)

        y = conv2d(output_filters, kernel_size=(1, 1), strides=(1, 1))(y)
        y = batch_normalization()(y)
        y = keras.layers.Activation(activation)(y)

        if use_skip:
            x = conv2d(output_filters, kernel_size=(1, 1), strides=strides)(x)
            x = batch_normalization()(x)

        y = keras.layers.Add()([y, x])
        y = keras.layers.Activation(activation)(y)
        return y

    return layer


def zblock(
    output_filters: int = 0,
    group_width: int = 0,
    kernel_size: int | tuple[int, int] = 3,
    strides: int | tuple[int, int] = 1,
    se_ratio: float = 4,
    activation: str = "relu6",
    name: str | None = None,
) -> Callable[[keras.KerasTensor], keras.KerasTensor]:
    """RegNet X-Block

    Args:
        output_filters (int, optional): Number of output filters. Defaults to 0.
        group_width (int, optional): Group width. Defaults to 0.
        kernel_size (int | tuple[int, int], optional): Kernel size. Defaults to 3.
        strides (int | tuple[int, int], optional): Stride length. Defaults to 1.
        se_ratio (float, optional): SE ratio. Defaults to 4.
        activation (str, optional): Activation function. Defaults to "relu6".
        name (str|None, optional): Block name. Defaults to None.

    Returns:
        Callable[[keras.KerasTensor], keras.KerasTensor]: Functional layer
    """

    def layer(x: keras.KerasTensor) -> keras.KerasTensor:
        input_filters = x.shape[-1]
        groups = input_filters // group_width
        use_add = input_filters == output_filters and (strides == 1 if isinstance(strides, int) else strides[0] == 1)
        expand_ratio = 2

        name_ex = f"{name}_exp" if name else None
        filters = input_filters * expand_ratio
        y = conv2d(filters, kernel_size=(1, 1), strides=(1, 1), name=name_ex)(x)
        y = batch_normalization(name=name_ex)(y)
        y = keras.layers.Activation(activation, name=f"{name_ex}_act")(y)

        y = keras.layers.Conv2D(
            filters,
            kernel_size=kernel_size,
            strides=strides,
            groups=groups,
            padding="same",
            use_bias=False,
            kernel_initializer=keras.initializers.VarianceScaling(),  # type: ignore
        )(y)
        y = batch_normalization()(y)
        y = keras.layers.Activation(activation)(y)

        if se_ratio:
            y = se_layer(ratio=se_ratio)(y)

        y = conv2d(output_filters, kernel_size=(1, 1), strides=(1, 1))(y)
        y = batch_normalization()(y)

        if use_add:
            y = keras.layers.Add()([y, x])

        return y

    return layer


def regnet_core(
    blocks: list[RegNetBlockParam],
    block_style: Literal["y", "z"] = "y",
) -> Callable[[keras.KerasTensor], keras.KerasTensor]:
    """RegNet core

    Args:
        blocks (list[RegNetBlockParam]): Block params
        block_style (float, optional): Block style. Defaults to 'y'.

    Returns:
        Callable[[keras.KerasTensor], keras.KerasTensor]: Core
    """

    def layer(x: keras.KerasTensor) -> keras.KerasTensor:
        block_fn = yblock if block_style == "y" else zblock
        for i, block in enumerate(blocks):
            filters = make_divisible(block.filters, 8)
            for d in range(block.depth):
                name = f"stage{i}_block{d + 1}"
                x = block_fn(
                    output_filters=filters,
                    group_width=block.group_width,
                    kernel_size=block.kernel_size,
                    strides=block.strides,
                    se_ratio=block.se_ratio,
                    activation=block.activation,
                    name=name,
                )(x)
            # END FOR
        # END FOR
        return x

    return layer


def regnet_layer(
    x: keras.KerasTensor,
    params: RegNetParams,
    num_classes: int | None = None,
) -> keras.KerasTensor:
    """Create RegNet TF functional model

    Args:
        x (keras.KerasTensor): Input tensor
        params (RegNetParams): Model parameters.
        num_classes (int, optional): Number of classes.

    Returns:
        keras.KerasTensor: Output tensor
    """
    requires_reshape = len(x.shape) == 3
    if requires_reshape:
        y = keras.layers.Reshape((1,) + x.shape[1:])(x)
    else:
        y = x
    # END IF

    # Stem
    if params.input_filters > 0:
        name = "stem"
        filters = make_divisible(params.input_filters, 8)
        y = conv2d(filters, kernel_size=(3, 3), strides=params.input_strides, name=name)(y)
        y = batch_normalization(name=name)(y)
        y = keras.layers.Activation(params.input_activation, name=f"{name}_act")(y)
    else:
        y = x

    y = regnet_core(blocks=params.blocks, block_style=params.block_style)(y)

    if params.output_filters:
        name = "neck"
        filters = make_divisible(params.output_filters, 8)
        y = conv2d(filters, kernel_size=(1, 1), strides=(1, 1), padding="valid", name=name)(y)
        y = batch_normalization(name=name)(y)
        y = keras.layers.Activation(params.output_activation, name=f"{name}_act")(y)

    if params.include_top:
        name = "top"
        y = keras.layers.GlobalAveragePooling2D(name=f"{name}_pool")(y)

        if params.dropout > 0 and params.dropout < 1:
            y = keras.layers.Dropout(params.dropout)(y)
        if num_classes is not None:
            y = keras.layers.Dense(num_classes, name=name)(y)
        if params.output_activation:
            y = keras.layers.Activation(params.output_activation)(y)

    # Only reshape if needed
    elif requires_reshape:
        y = keras.layers.Reshape(y.shape[2:])(y)

    return y


class RegNetModel:
    """Helper class to generate model from parameters"""

    @staticmethod
    def layer_from_params(inputs: keras.Input, params: RegNetParams | dict, num_classes: int | None = None):
        """Create layer from parameters"""
        if isinstance(params, dict):
            params = RegNetParams(**params)
        return regnet_layer(x=inputs, params=params, num_classes=num_classes)

    @staticmethod
    def model_from_params(inputs: keras.Input, params: RegNetParams | dict, num_classes: int | None = None):
        """Create model from parameters"""
        outputs = RegNetModel.layer_from_params(inputs=inputs, params=params, num_classes=num_classes)
        return keras.Model(inputs=inputs, outputs=outputs)



"""
# Random Background Noises Layer API

This module provides classes to build random background noises layers.

Classes:
    RandomBackgroundNoises1D: Random background noises 1D


"""

import keras

from .base_augmentation import BaseAugmentation1D
from ...utils import parse_factor, helia_export


@helia_export(path="helia_edge.layers.preprocessing.RandomBackgroundNoises1D")
class RandomBackgroundNoises1D(BaseAugmentation1D):
    amplitude: tuple[float, float]
    num_noises: int

    def __init__(self, noises, amplitude: float | tuple[float, float] = 0.1, num_noises: int = 1, **kwargs):
        """Apply random background noises to the input.

        Args:
            noises (np.ndarray): Background noises to apply.
            amplitude (float|tuple[float,float]): Amplitude of the noise. If tuple, amplitude is randomly picked between the values.

        Example:

        ```python
            sample_rate = 100
            duration = 2*sample_rate
            freqs = [2, 5, 15, 20, 25]
            noises = np.vstack([
                np.sin(2*np.pi*f*np.arange(duration)/sample_rate)
                for f in freqs
            ]).T
            lyr = RandomBackgroundNoises(noises=noises, amplitude=0.2, num_noises=2)
            y = lyr(x, training=True)
        ```
        """
        super().__init__(**kwargs)

        self.amplitude = parse_factor(amplitude, min_value=0, max_value=None, param_name="amplitude")
        self.num_noises = num_noises
        self.noises = noises

    def get_random_transformations(self, input_shape: tuple[int, int, int]) -> dict:
        """Generate noise tensor

        Args:
            input_shape (tuple[int, ...]): Input shape.

        Returns:
            dict: Dictionary containing the noise tensor.
        """
        batch_size = input_shape[0]
        duration_size = input_shape[self.data_axis]

        noise_idx = keras.random.randint(
            shape=(batch_size, self.num_noises),
            minval=0,
            maxval=self.noises.shape[1],
        )
        start = keras.random.randint(
            shape=(batch_size, self.num_noises),
            minval=0,
            maxval=self.noises.shape[0] - duration_size + 1,
            seed=self.random_generator,
            dtype="int32",
        )
        amplitude = keras.random.uniform(
            shape=(batch_size, self.num_noises),
            minval=self.amplitude[0],
            maxval=self.amplitude[1],
            seed=self.random_generator,
            dtype=self.compute_dtype,
        ) / keras.ops.cast(self.num_noises, self.compute_dtype)

        return {
            "noise_idx": noise_idx,
            "start": start,
            "amplitude": amplitude,
        }

    def augment_sample(self, inputs) -> keras.KerasTensor:
        """Augment single sample with random background noises."""
        duration_size = inputs[self.SAMPLES].shape[self.data_axis]
        ch_size = inputs[self.SAMPLES].shape[self.ch_axis]

        def random_noise(i, x):
            noise_idx = inputs[self.TRANSFORMS]["noise_idx"][i]
            start = inputs[self.TRANSFORMS]["start"][i]
            amplitude = inputs[self.TRANSFORMS]["amplitude"][i]
            noise = keras.ops.slice(self.noises, (start, noise_idx), (duration_size, 1))
            noise = keras.ops.squeeze(noise)
            if self.data_format == "channels_first":
                noise = keras.ops.reshape(noise, (1, duration_size))
                noise = keras.ops.tile(noise, (ch_size, 1))
            else:
                noise = keras.ops.reshape(noise, (duration_size, 1))
                noise = keras.ops.tile(noise, (1, ch_size))
            return x + amplitude * noise

        # END DEF

        if self.training:
            sample = inputs[self.SAMPLES]
            outputs = keras.ops.fori_loop(lower=0, upper=self.num_noises, body_fun=random_noise, init_val=sample)
        else:
            outputs = inputs[self.SAMPLES]
        return outputs

    def get_config(self):
        """Serializes the configuration of the layer."""
        config = super().get_config()
        config.update(
            {
                "amplitude": self.amplitude,
                "num_noises": self.num_noises,
            }
        )
        return config



"""
# :material-layers: Layers API

The `helia_edge.layers` module provides classes to build neural network layers.
For example, you can use the `helia_edge.layers.preprocessing.AmplitudeWarp` layer to apply amplitude warping to audio signals.

## Available Layers

* **[Preprocessing Layers](./preprocessing)**: Provides `tf.data.Dataset` preprocessing layers.
* **[Activations](./activations)**: Provides activation functions.
* **[Convolutional Layers](./convolutional)**: Provides convolutional layers.
* **[Gumbel Softmax Bottleneck](./gumbel_softmax_bottleneck)**: Provides Gumbel Softmax bottleneck layer.
* **[Normalization Layers](./normalization)**: Provides normalization layers.
* **[Patching Layers](./patching)**: Provides patching layers.
* **[Residual Vector Quantizer](./residual_vector_quantizer)**: Provides residual vector quantizer layer.
* **[Squeeze-and-Excitation Layer](./squeeze_excite)**: Provides squeeze-and-excitation layers.
* **[MBConv Layer](./mbconv)**: Provides popular mbconv block.
* **[Vector Quantizer](./vector_quantizer)**: Provides vector quantizer layer.


"""

from . import preprocessing
from . import activations
from . import convolutional
from . import mbconv
from . import normalization
from . import patching
from . import squeeze_excite

from .activations import swish, glu, relu, relu6, sigmoid, mish, gelu
from .convolutional import conv1d, conv2d
from .gumbel_softmax_bottleneck import GumbelSoftmaxBottleneck
from .mbconv import mbconv_block, MBConvParams
from .normalization import batch_normalization, layer_normalization
from .patching import PatchLayer2D, MaskedPatchEncoder2D
from .residual_vector_quantizer import ResidualVectorQuantizer
from .squeeze_excite import se_layer
from .vector_quantizer import VectorQuantizer



"""
# Contrastive Trainer API

This module provides a trainer for contrastive learning.

Classes:
    ContrastiveTrainer: Trainer for contrastive learning

"""

import keras
import tensorflow as tf
from ..utils import convert_inputs_to_tf_dataset, helia_export


@helia_export(path="helia_edge.trainers.ContrastiveTrainer")
class ContrastiveTrainer(keras.Model):
    SAMPLES = "data"
    LABELS = "labels"
    AUG_SAMPLES_0 = "augmented_data_0"
    AUG_SAMPLES_1 = "augmented_data_1"

    encoder: keras.Model
    augmenters: tuple[keras.Layer, keras.Layer]

    def __init__(
        self,
        encoder: keras.Model,
        projector: keras.Model | tuple[keras.Model, keras.Model],
        augmenter: keras.Layer | tuple[keras.Layer, keras.Layer] | None = None,
        probe: keras.Layer | keras.Model | None = None,
    ):
        """Creates a self-supervised contrastive trainer for a model.

        Args:
            encoder (keras.Model): The encoder model to be trained.
            projector (keras.Model|tuple[keras.Model, keras.Model]): The projector model to be trained.
            augmenter (keras.Layer|tuple[keras.Layer, keras.Layer]|None): The augmenter to be used for data augmentation.
            probe (keras.Layer|keras.Model|None): The probe model to be trained. If None, no probe is used.

        """
        super().__init__()

        if len(encoder.output.shape) != 2:
            raise ValueError(
                f"`encoder` must have a flattened output. Expected "
                f"rank(encoder.output.shape)=2, got "
                f"encoder.output.shape={encoder.output.shape}"
            )

        if isinstance(augmenter, tuple) and len(augmenter) != 2:
            raise ValueError("`augmenter` must be either a single augmenter or a tuple of exactly 2 augmenters.")

        if isinstance(projector, tuple) and len(projector) != 2:
            raise ValueError("`projector` must be either a single augmenter or a tuple of exactly 2 augmenters.")

        if augmenter is None:
            self.augmenters = (keras.layers.Lambda(lambda x: x), keras.layers.Lambda(lambda x: x))
        elif isinstance(augmenter, tuple):
            self.augmenters = augmenter
        else:
            self.augmenters = (augmenter, augmenter)

        self.encoder = encoder

        # Check to see if the projector is being shared or are distinct.
        self._is_shared_projector = True if not isinstance(projector, tuple) else False
        self.projectors = projector if type(projector) is tuple else (projector, projector)
        self.probe = probe

        self.loss_metric = keras.metrics.Mean(name="loss")
        self.encoder_metrics = []
        if probe is not None:
            self.probe_loss_metric = keras.metrics.Mean(name="probe_loss")
            self.probe_metrics = []

    def compile(
        self,
        encoder_optimizer: keras.Optimizer,
        encoder_loss: keras.Loss,
        encoder_metrics: list[keras.Metric] | None = None,
        probe_optimizer: keras.Optimizer | None = None,
        probe_loss: keras.Loss | None = None,
        probe_metrics: list[keras.Metric] | None = None,
        **kwargs,
    ):
        super().compile(
            loss=encoder_loss,
            optimizer=encoder_optimizer,
            metrics=None,
            **kwargs,
        )

        if self.probe and not probe_optimizer:
            raise ValueError("`probe_optimizer` must be specified when a probe is included.")

        if self.probe and not probe_loss:
            raise ValueError("`probe_loss` must be specified when a probe is included.")

        if "loss" in kwargs:
            raise ValueError(
                "`loss` parameter in ContrastiveTrainer.compile is ambiguous. "
                "Please specify `encoder_loss` or `probe_loss`."
            )

        if "optimizer" in kwargs:
            raise ValueError(
                "`optimizer` parameter in ContrastiveTrainer.compile is "
                "ambiguous. Please specify `encoder_optimizer` or "
                "`probe_optimizer`."
            )

        if "metrics" in kwargs:
            raise ValueError(
                "`metrics` parameter in ContrastiveTrainer.compile is "
                "ambiguous. Please specify `encoder_metrics` or "
                "`probe_metrics`."
            )
        self.encoder_metrics = encoder_metrics or []
        if self.probe:
            self.probe_loss = probe_loss
            self.probe_optimizer = probe_optimizer
            self.probe_metrics = probe_metrics or []

    @property
    def metrics(self):
        metrics = [self.loss_metric]
        if self.encoder_metrics:
            metrics += self.encoder_metrics
        if self.probe:
            metrics += [self.probe_loss_metric]
            metrics += self.probe_metrics
        return metrics

    def fit(
        self,
        x=None,
        y=None,
        sample_weight=None,
        batch_size=None,
        validation_data=None,
        **kwargs,
    ):
        # Force training to tf.data and apply augmentations
        train_ds = (
            convert_inputs_to_tf_dataset(x=x, y=y, sample_weight=sample_weight, batch_size=batch_size)
            .map(self.run_augmenters, num_parallel_calls=tf.data.AUTOTUNE)
            .prefetch(tf.data.AUTOTUNE)
        )

        # If validation data is provided, apply augmentations
        if validation_data:
            val_ds = (
                convert_inputs_to_tf_dataset(
                    x=validation_data,
                    batch_size=batch_size,
                )
                .map(self.run_augmenters, num_parallel_calls=tf.data.AUTOTUNE)
                .prefetch(tf.data.AUTOTUNE)
            )
        else:
            val_ds = None
        # END IF

        return super().fit(x=train_ds, validation_data=val_ds, **kwargs)

    def run_augmenters(self, x, y=None):
        if isinstance(x, dict):
            inputs = x
        else:
            inputs = {self.SAMPLES: x}

        if y is not None:
            inputs[self.LABELS] = y

        if self.AUG_SAMPLES_0 not in inputs:
            inputs[self.AUG_SAMPLES_0] = self.augmenters[0](x, training=True)
        if self.AUG_SAMPLES_1 not in inputs:
            inputs[self.AUG_SAMPLES_1] = self.augmenters[1](x, training=True)

        return inputs

    def _tensorflow_train_step(self, data):
        samples = data[self.SAMPLES]
        labels = data[self.LABELS] if self.LABELS in data else None
        augmented_samples_0 = data[self.AUG_SAMPLES_0]
        augmented_samples_1 = data[self.AUG_SAMPLES_1]

        with tf.GradientTape() as tape:
            features_0 = self.encoder(augmented_samples_0, training=True)
            features_1 = self.encoder(augmented_samples_1, training=True)

            projections_0 = self.projectors[0](features_0, training=True)
            projections_1 = self.projectors[1](features_1, training=True)

            loss = self.compiled_loss(
                projections_0,
                projections_1,
                regularization_losses=self.encoder.losses,
            )

        # If the projector is shared, then take the trainable weights of just
        # one of the projectors in the tuple. If not, use both the projectors.
        projector_weights = (
            self.projectors[0].trainable_weights
            if self._is_shared_projector
            else self.projectors[0].trainable_weights + self.projectors[1].trainable_weights
        )
        gradients = tape.gradient(
            loss,
            self.encoder.trainable_weights + projector_weights,
        )

        self.optimizer.apply_gradients(
            zip(
                gradients,
                self.encoder.trainable_weights + projector_weights,
            )
        )

        # Update the metrics
        self.loss_metric.update_state(loss)
        for metric in self.encoder_metrics:
            metric.update_state(features_0, features_1)

        if self.probe:
            if labels is None:
                raise ValueError("Targets must be provided when a probe is specified")
            with tf.GradientTape() as tape:
                features = tf.stop_gradient(self.encoder(samples, training=False))
                class_logits = self.probe(features, training=True)
                probe_loss = self.probe_loss(labels, class_logits)
            gradients = tape.gradient(probe_loss, self.probe.trainable_weights)
            self.probe_optimizer.apply_gradients(zip(gradients, self.probe.trainable_weights))
            self.probe_loss_metric.update_state(probe_loss)
            for metric in self.probe_metrics:
                metric.update_state(labels, class_logits)

        return {metric.name: metric.result() for metric in self.metrics}

    def _tensorflow_test_step(self, data):
        # Called by fit
        if isinstance(data, dict):
            labels = data[self.LABELS] if self.LABELS in data else None
            augmented_samples_0 = data[self.AUG_SAMPLES_0]
            augmented_samples_1 = data[self.AUG_SAMPLES_1]
        # Called by evaluate (need to compute augmentations)
        else:
            samples = data
            augmented_samples_0 = self.augmenters[0](samples, training=True)
            augmented_samples_1 = self.augmenters[1](samples, training=True)
            labels = None
        # END IF

        features_0 = self.encoder(augmented_samples_0, training=True)
        features_1 = self.encoder(augmented_samples_1, training=True)

        projections_0 = self.projectors[0](features_0, training=True)
        projections_1 = self.projectors[1](features_1, training=True)

        loss = self.compiled_loss(
            projections_0,
            projections_1,
            regularization_losses=self.encoder.losses,
        )

        # Update the metrics
        self.loss_metric.update_state(loss)
        for metric in self.encoder_metrics:
            metric.update_state(features_0, features_1)

        if self.probe:
            class_logits = self.probe(features_0, training=False)
            probe_loss = self.probe_loss(labels, class_logits)
            self.probe_loss_metric.update_state(probe_loss)
            for metric in self.probe_metrics:
                metric.update_state(labels, class_logits)
        return {metric.name: metric.result() for metric in self.metrics}

    def train_step(self, data):
        if keras.backend.backend() == "tensorflow":
            return self._tensorflow_train_step(data)
        elif keras.backend.backend() == "jax":
            raise NotImplementedError("JAX backend is not supported.")
        elif keras.backend.backend() == "torch":
            raise NotImplementedError("PyTorch backend is not supported.")

    def test_step(self, data):
        if keras.backend.backend() == "tensorflow":
            return self._tensorflow_test_step(data)
        elif keras.backend.backend() == "jax":
            raise NotImplementedError("JAX backend is not supported.")
        elif keras.backend.backend() == "torch":
            raise NotImplementedError("PyTorch backend is not supported.")

    def call(self, inputs):
        raise NotImplementedError("ContrastiveTrainer.call() is not implemented - please call your model directly.")

    @staticmethod
    def linear_probe(num_classes, **kwargs):
        return keras.Sequential(keras.layers.Dense(num_classes), **kwargs)

    def save(self, filepath, overwrite=True, zipped=True, **kwargs):
        """We only save the encoder model"""
        self.encoder.save(filepath, overwrite=overwrite, zipped=zipped, **kwargs)


# class MomentumContrastiveTrainer(ContrastiveTrainer):
#     """Base class for momentum contrastive learning models"""

#     def __init__(
#         self,
#         encoder: keras.Model,
#         projector: keras.Model,
#         contrastive_augmenter: Callable[[tf.Tensor], tf.Tensor] | None = None,
#         classification_augmenter: Callable[[tf.Tensor], tf.Tensor] | None = None,
#         linear_probe: keras.Model | None = None,
#         momentum_coeff: float = 0.999,
#     ):
#         super().__init__(
#             encoder=encoder,
#             projector=projector,
#             contrastive_augmenter=contrastive_augmenter,
#             classification_augmenter=classification_augmenter,
#             linear_probe=linear_probe,
#         )
#         self.momentum_coeff = momentum_coeff

#         # the momentum networks are initialized from their online counterparts
#         self.m_encoder = keras.models.clone_model(self.encoder)
#         self.m_projector = keras.models.clone_model(self.projector)

#     @abstractmethod
#     def contrastive_loss(
#         self,
#         projections_1,
#         projections_2,
#         m_projections_1,
#         m_projections_2,
#     ):  # pylint: disable=arguments-differ
#         pass

#     def train_step(self, data):
#         """Training step for the model"""
#         pair1, pair2 = data

#         # each input is augmented twice, differently
#         augmented_inputs_1 = self.contrastive_augmenter(pair1)
#         augmented_inputs_2 = self.contrastive_augmenter(pair2)

#         with tf.GradientTape() as tape:
#             # Encoder phase
#             features_1 = self.encoder(augmented_inputs_1)
#             features_2 = self.encoder(augmented_inputs_2)
#             # Projection phase
#             projections_1 = self.projector(features_1)
#             projections_2 = self.projector(features_2)
#             # Momentum encoder phase
#             m_features_1 = self.m_encoder(augmented_inputs_1)
#             m_features_2 = self.m_encoder(augmented_inputs_2)
#             # Momentum projection phase
#             m_projections_1 = self.m_projector(m_features_1)
#             m_projections_2 = self.m_projector(m_features_2)
#             contrastive_loss = self.contrastive_loss(projections_1, projections_2, m_projections_1, m_projections_2)
#         # END WITH

#         # backpropagation
#         gradients = tape.gradient(
#             contrastive_loss,
#             self.encoder.trainable_weights + self.projector.trainable_weights,
#         )
#         self.contrastive_optimizer.apply_gradients(
#             zip(
#                 gradients,
#                 self.encoder.trainable_weights + self.projector.trainable_weights,
#             )
#         )
#         self.contrastive_loss_tracker.update_state(contrastive_loss)
#         self.correlation_loss(m_features_1, m_features_2)

#         self.update_contrastive_accuracy(m_features_1, m_features_2)
#         self.update_correlation_accuracy(m_features_1, m_features_2)

#         # labeled_inputs = None
#         # labels = None
#         # preprocessed_inputs = self.classification_augmenter(labeled_inputs)
#         # with tf.GradientTape() as tape:
#         #     # the momentum encoder is used here as it moves more slowly
#         #     features = self.m_encoder(preprocessed_inputs)
#         #     class_logits = self.linear_probe(features)
#         #     probe_loss = self.probe_loss(labels, class_logits)
#         # gradients = tape.gradient(probe_loss, self.linear_probe.trainable_weights)
#         # self.probe_optimizer.apply_gradients(
#         #     zip(gradients, self.linear_probe.trainable_weights)
#         # )
#         # self.probe_accuracy.update_state(labels, class_logits)

#         # the momentum networks are updated by exponential moving average
#         for weight, m_weight in zip(self.encoder.weights, self.m_encoder.weights):
#             m_weight.assign(self.momentum_coeff * m_weight + (1 - self.momentum_coeff) * weight)
#         for weight, m_weight in zip(self.projector.weights, self.m_projector.weights):
#             m_weight.assign(self.momentum_coeff * m_weight + (1 - self.momentum_coeff) * weight)

#         return {m.name: m.result() for m in self.metrics}

#     def test_step(self, data):
#         """Test step for the model"""
#         pair1, pair2 = data
#         augmented_inputs_1 = self.contrastive_augmenter(pair1)
#         augmented_inputs_2 = self.contrastive_augmenter(pair2)
#         features_1 = self.encoder(augmented_inputs_1, training=False)
#         features_2 = self.encoder(augmented_inputs_2, training=False)
#         projections_1 = self.projector(features_1, training=False)
#         projections_2 = self.projector(features_2, training=False)
#         m_features_1 = self.m_encoder(augmented_inputs_1, training=False)
#         m_features_2 = self.m_encoder(augmented_inputs_2, training=False)
#         m_projections_1 = self.m_projector(m_features_1, training=False)
#         m_projections_2 = self.m_projector(m_features_2, training=False)

#         contrastive_loss = self.contrastive_loss(projections_1, projections_2, m_projections_1, m_projections_2)

#         self.contrastive_loss_tracker.update_state(contrastive_loss)
#         self.correlation_loss(m_features_1, m_features_2)

#         self.update_contrastive_accuracy(m_features_1, m_features_2)
#         self.update_correlation_accuracy(m_features_1, m_features_2)

#         return {m.name: m.result() for m in self.metrics}


# class MoCo(MomentumContrastiveModel):
#     """MoCo model for self-supervised learning"""

#     def __init__(
#         self,
#         encoder: keras.Model,
#         projector: keras.Model,
#         contrastive_augmenter: Callable[[tf.Tensor], tf.Tensor] | None = None,
#         classification_augmenter: Callable[[tf.Tensor], tf.Tensor] | None = None,
#         linear_probe: keras.Model | None = None,
#         momentum_coeff: float = 0.999,
#         temperature: float = 0.1,
#         queue_size: int = 65536,
#     ):
#         super().__init__(
#             encoder=encoder,
#             projector=projector,
#             contrastive_augmenter=contrastive_augmenter,
#             classification_augmenter=classification_augmenter,
#             linear_probe=linear_probe,
#             momentum_coeff=momentum_coeff,
#         )
#         self.temperature = temperature

#         feature_dimensions = encoder.output_shape[1]
#         self.feature_queue = tf.Variable(
#             tf.math.l2_normalize(tf.random.normal(shape=(queue_size, feature_dimensions)), axis=1),
#             trainable=False,
#         )

#     def contrastive_loss(
#         self,
#         projections_1,
#         projections_2,
#         m_projections_1,
#         m_projections_2,
#     ):
#         # similar to the SimCLR loss, however it uses the momentum networks'
#         # representations of the differently augmented views as targets
#         projections_1 = tf.math.l2_normalize(projections_1, axis=1)
#         projections_2 = tf.math.l2_normalize(projections_2, axis=1)
#         m_projections_1 = tf.math.l2_normalize(m_projections_1, axis=1)
#         m_projections_2 = tf.math.l2_normalize(m_projections_2, axis=1)

#         similarities_1_2 = (
#             tf.matmul(
#                 projections_1,
#                 tf.concat((m_projections_2, self.feature_queue), axis=0),
#                 transpose_b=True,
#             )
#             / self.temperature
#         )
#         similarities_2_1 = (
#             tf.matmul(
#                 projections_2,
#                 tf.concat((m_projections_1, self.feature_queue), axis=0),
#                 transpose_b=True,
#             )
#             / self.temperature
#         )

#         batch_size = tf.shape(projections_1)[0]
#         contrastive_labels = tf.range(batch_size)
#         loss = keras.losses.sparse_categorical_crossentropy(
#             tf.concat([contrastive_labels, contrastive_labels], axis=0),
#             tf.concat([similarities_1_2, similarities_2_1], axis=0),
#             from_logits=True,
#         )

#         # feature queue update
#         self.feature_queue.assign(
#             tf.concat(
#                 [
#                     m_projections_1,
#                     m_projections_2,
#                     self.feature_queue[: -(2 * batch_size)],
#                 ],
#                 axis=0,
#             )
#         )
#         return loss



"""
# MetaFormer: Meta-Learning with Transformers

## Overview

MetaFormer is a transformer-based model that incorporates both spatial mixing and channel mixing blocks.
The architecture is designed to learn from few examples and generalize to new tasks.

For more info, refer to the original paper [MetaFormer: Meta-Learning with Transformers](https://arxiv.org/abs/2110.11605).

Classes:
    MetaFormerParams: MetaFormer parameters
    MetaFormerModel: Helper class to generate model from parameters

Functions:
    patch_embedding: Patch embedding layer
    pool_token_mixer: Token mixer using average pooling
    conv_token_mixer: Token mixer using separable convolution
    attention_token_mixer: Token mixer using multi-head attention
    mlp_channel_mixer: Channel mixer using MLP via 1x1 convolutions
    metaformer_block: Metaformer block
    metaformer_layer: Metaformer functional layer

"""

import keras
from pydantic import BaseModel, Field


class NameArgs(BaseModel):
    """Name and arguments

    Attributes:
        name (str): Name
        args (dict): Arguments

    """

    name: str = Field(default="conv", description="Name")
    args: dict = Field(default_factory=dict, description="Arguments")


class MetaFormerBlockParams(BaseModel):
    """MetaFormer block parameters

    Attributes:
        layers (int): Number of layers
        patch_embed (dict): Patch embedding
        token_mixer (NameArgs): Token mixer
        channel_mixer (NameArgs): Channel mixer

    """

    layers: int = Field(default=2, description="Number of layers")
    patch_embed: dict = Field(default_factory=dict, description="Patch embedding")
    token_mixer: NameArgs = Field(default_factory=dict, description="Token mixer")
    channel_mixer: NameArgs = Field(default_factory=dict, description="Channel mixer")


class MetaFormerParams(BaseModel):
    """MetaFormer parameters

    Attributes:
        blocks (list[MetaFormerBlockParams]): MetaFormer blocks
        output_filters (int): Output filters
        output_activation (str | None): Output activation
        include_top (bool): Include top
        dropout (float): Dropout rate
        drop_connect_rate (float): Drop connect rate
        name (str): Model name

    """

    blocks: list[MetaFormerBlockParams] = Field(default_factory=list, description="MetaFormer blocks")
    output_filters: int = Field(default=0, description="Output filters")
    output_activation: str | None = Field(default=None, description="Output activation")
    include_top: bool = Field(default=True, description="Include top")
    dropout: float = Field(default=0.2, description="Dropout rate")
    drop_connect_rate: float = Field(default=0.2, description="Drop connect rate")
    name: str = Field(default="MetaFormer", description="Model name")


def patch_embedding(
    embed_dim: int,
    patch_shape: tuple[int, int],
    stride_shape: tuple[int, int] | None = None,
    padding: str = "same",
) -> keras.layers.Layer:
    """Patch embedding layer using 2D convolution

    Args:
        embed_dim (int): Embedding dimension
        patch_shape (tuple[int, int]): Patch shape
        stride_shape (tuple[int, int], optional): Stride shape. Defaults to None.
        padding (str, optional): Padding. Defaults to 'same'.

    """
    # SHAPE (B, F, T, C) -> (B, F//S, T//S, E)
    # SHAPE (B, H, W, C) -> (B, H//S, W//S, E
    return keras.layers.Conv2D(
        filters=embed_dim,
        kernel_size=patch_shape,
        strides=stride_shape or patch_shape,
        padding=padding,
        use_bias=False,
    )


def pool_token_mixer(
    pool_size: tuple[int, int] = (2, 2),
) -> keras.layers.Layer:
    """Token mixer using average pooling

    Args:
        pool_size (tuple[int, int], optional): Pool size. Defaults to (2, 2).

    Returns:
        keras.layers.Layer: Token mixer layer
    """

    def layer(x: keras.KerasTensor) -> keras.KerasTensor:
        y = x
        y = (
            keras.layers.AveragePooling2D(
                pool_size=pool_size,
                strides=(1, 1),
                padding="same",
            )(y)
            - y
        )
        return y

    # END DEF
    return layer


def conv_token_mixer(
    embed_dim: int,
    kernel_size: tuple[int, int] = (3, 3),
    strides: tuple[int, int] = (1, 1),
) -> keras.Layer:
    """Token mixer using separable convolution

    Args:
        embed_dim (int): Embedding dimension
        kernel_size (tuple[int, int], optional): Kernel size. Defaults to (3, 3).
        strides (tuple[int, int], optional): Strides. Defaults to (1, 1).

    Returns:
        keras.layers.Layer: Token mixer layer
    """

    def layer(x: keras.KerasTensor) -> keras.KerasTensor:
        y = x
        y = keras.layers.SeparableConv2D(
            filters=embed_dim,
            kernel_size=kernel_size,
            strides=strides,
            padding="same",
            use_bias=False,
        )(y)
        return y

    # END DEF
    return layer


def attention_token_mixer(
    embed_dim: int,
    num_heads: int,
    dropout: float = 0.1,
) -> keras.layers.Layer:
    """Token mixer using multi-head attention

    Args:
        embed_dim (int): Embedding dimension
        num_heads (int): Number of heads
        dropout (float, optional): Dropout rate. Defaults to 0.1.

    Returns:
        keras.layers.Layer: Token mixer layer
    """

    def layer(x: keras.KerasTensor) -> keras.KerasTensor:
        input_shape = keras.ops.shape(x)
        height = input_shape[1]
        width = input_shape[2]
        channels = input_shape[3]
        y = x  # Shape is B, H, W, C
        y = keras.layers.Reshape((height * width, channels))(y)
        y = keras.layers.MultiHeadAttention(
            num_heads=num_heads,
            key_dim=embed_dim,
            dropout=dropout,
        )(y, y)
        # Reshape back to B, H, W, C
        y = keras.layers.Reshape((height, width, channels))(y)
        return y

    # END DEF
    return layer


def mlp_channel_mixer(
    embed_dim: int,
    ratio: int = 4,
    activation: str = "gelu",
    dropout: float = 0,
) -> keras.Layer:
    """Channel mixer using MLP via 1x1 convolutions

    Args:
        embed_dim (int): Embedding dimension
        ratio (int, optional): Expansion ratio. Defaults to 4.
        activation (str, optional): Activation function. Defaults to "gelu".
        dropout (float, optional): Dropout rate. Defaults to 0.

    Returns:
        keras.layers.Layer: Channel mixer layer
    """

    def layer(x: keras.KerasTensor) -> keras.KerasTensor:
        y = x
        y = keras.layers.Conv2D(
            filters=int(embed_dim * ratio),
            kernel_size=(1, 1),
            activation=activation,
        )(y)

        if 0 < dropout < 1:
            y = keras.layers.Dropout(dropout)(y)

        y = keras.layers.Conv2D(
            filters=embed_dim,
            kernel_size=(1, 1),
        )(y)

        if 0 < dropout < 1:
            y = keras.layers.Dropout(dropout)(y)

        return y

    # END DEF
    return layer


def metaformer_block(
    token_mixer: keras.layers.Layer | None = None,
    channel_mixer: keras.layers.Layer | None = None,
    name: str = "mf_block",
) -> keras.layers.Layer:
    """Metaformer block

    Args:
        token_mixer (keras.layers.Layer, optional): Token mixer layer. Defaults to None.
        channel_mixer (keras.layers.Layer, optional): Channel mixer layer. Defaults to None.
        name (str, optional): Block name. Defaults to 'mf_block'.

    Returns:
        keras.layers.Layer: Metaformer block
    """

    def layer(x: keras.KerasTensor) -> keras.KerasTensor:
        # Shape is B, T, F, C
        y = x

        if token_mixer:
            # Apply layer normalization
            y1 = keras.layers.LayerNormalization(name=f"{name}_token_ln", axis=-1)(y)

            # Apply token mixer
            y1 = token_mixer(y1)

            # Add residual connection
            y = y + y1
        # END IF

        if channel_mixer:
            # Apply layer normalization
            y1 = keras.layers.LayerNormalization(name=f"{name}_ch_ln", axis=-1)(y)

            # Apply channel mixer
            y1 = channel_mixer(y1)

            # Add residual connection
            y = y + y1

        return y

    # END DEF

    return layer


def metaformer_layer(
    x: keras.KerasTensor,
    params: MetaFormerParams,
    num_classes: int | None = None,
) -> keras.KerasTensor:
    """MetaFormer functional layer

    Args:
        x (keras.KerasTensor): Input tensor
        params (MetaFormerParams): Model parameters.
        num_classes (int, optional): Number of classes.

    Returns:
        keras.KerasTensor: Output tensor
    """
    y = x

    for b, block in enumerate(params.blocks):
        print("Adding block", b + 1)
        # Apply patch embedding
        y = patch_embedding(**block.patch_embed)(y)
        for lyr in range(block.layers):
            token_mixer = None
            if block.token_mixer.name == "conv":
                token_mixer = conv_token_mixer(**block.token_mixer.args)
            elif block.token_mixer.name == "pool":
                token_mixer = pool_token_mixer(**block.token_mixer.args)
            elif block.token_mixer.name == "attention":
                token_mixer = attention_token_mixer(**block.token_mixer.args)
            # END IF
            channel_mixer = None
            if block.channel_mixer.name == "mlp":
                channel_mixer = mlp_channel_mixer(**block.channel_mixer.args)
            # END IF
            y = metaformer_block(
                token_mixer=token_mixer,
                channel_mixer=channel_mixer,
                name=f"B{b + 1}_L{lyr + 1}",
            )(y)
        # END FOR
    # END FOR

    if params.include_top:
        name = "top"
        y = keras.layers.GlobalAveragePooling2D(name=f"{name}_gap")(y)
        if 0 < params.dropout < 1:
            y = keras.layers.Dropout(params.dropout, name=f"{name}_dropout")(y)
        if num_classes is not None:
            y = keras.layers.Dense(num_classes, name=name)(y)
        if params.output_activation:
            y = keras.layers.Activation(params.output_activation)(y)
    # END IF


# def ccaa_metaformer(
#     x: keras.KerasTensor,
#     num_classes: int | None = None,
# ) -> keras.Model:
#     """CCAA Metaformer model"""

#     params = MetaFormerParams(
#         blocks=[
#             MetaFormerBlockParams(
#                 layers=2,
#                 patch_embed=dict(
#                     embed_dim=32,
#                     patch_shape=(7, 7),
#                     stride_shape=(4, 4),
#                 ),
#                 token_mixer=dict(
#                     name="conv",
#                     args=dict(
#                         embed_dim=32,
#                         kernel_size=(3, 3),
#                         strides=(1, 1),
#                     ),
#                 ),
#                 channel_mixer=dict(
#                     name="mlp",
#                     args=dict(
#                         embed_dim=32,
#                         ratio=4,
#                         activation="gelu",
#                         dropout=0.1,
#                     ),
#                 ),
#             ),
#             MetaFormerBlockParams(
#                 layers=2,
#                 patch_embed=dict(
#                     embed_dim=64,
#                     patch_shape=(3, 3),
#                     stride_shape=(2, 2),
#                 ),
#                 token_mixer=dict(
#                     name="conv",
#                     args=dict(
#                         embed_dim=64,
#                         kernel_size=(3, 3),
#                         strides=(1, 1),
#                     ),
#                 ),
#                 channel_mixer=dict(
#                     name="mlp",
#                     args=dict(
#                         embed_dim=64,
#                         ratio=4,
#                         activation="gelu",
#                         dropout=0.1,
#                     ),
#                 ),
#             ),
#             MetaFormerBlockParams(
#                 layers=2,
#                 patch_embed=dict(
#                     embed_dim=128,
#                     patch_shape=(3, 3),
#                     stride_shape=(2, 2),
#                 ),
#                 token_mixer=dict(
#                     name="conv",
#                     args=dict(
#                         embed_dim=128,
#                         kernel_size=(3, 3),
#                         strides=(1, 1),
#                     ),
#                 ),
#                 channel_mixer=dict(
#                     name="mlp",
#                     args=dict(
#                         embed_dim=128,
#                         ratio=4,
#                         activation="gelu",
#                         dropout=0.1,
#                     ),
#                 ),
#             ),
#             MetaFormerBlockParams(
#                 layers=2,
#                 patch_embed=dict(
#                     embed_dim=256,
#                     patch_shape=(3, 2),
#                     stride_shape=(3, 2),
#                 ),
#                 token_mixer=dict(
#                     name="conv",
#                     args=dict(
#                         embed_dim=256,
#                         kernel_size=(3, 3),
#                         strides=(1, 1),
#                     ),
#                 ),
#                 channel_mixer=dict(
#                     name="mlp",
#                     args=dict(
#                         embed_dim=256,
#                         ratio=4,
#                         activation="gelu",
#                         dropout=0.1,
#                     ),
#                 ),
#             ),
#         ],
#         include_top=True,
#         output_activation="softmax",
#     )

#     return metaformer_layer(
#         x=x,
#         params=params,
#         num_classes=num_classes,
#     )


class MetaFormerModel:
    """Helper class to generate model from parameters"""

    @staticmethod
    def layer_from_params(inputs: keras.Input, params: MetaFormerParams | dict, num_classes: int | None = None):
        """Create layer from parameters"""
        if isinstance(params, dict):
            params = MetaFormerParams(**params)
        return metaformer_layer(x=inputs, params=params, num_classes=num_classes)

    @staticmethod
    def model_from_params(inputs: keras.Input, params: MetaFormerParams | dict, num_classes: int | None = None):
        """Create model from parameters"""
        outputs = MetaFormerModel.layer_from_params(inputs=inputs, params=params, num_classes=num_classes)
        return keras.Model(inputs=inputs, outputs=outputs)
